import collections
import typing

import more_itertools as mit
import numpy as np
import opytional as opyt
import pandas as pd

from ...._auxiliary_lib import (
    HereditaryStratigraphicArtifact,
    alifestd_collapse_unifurcations,
    alifestd_count_children_of_asexual,
    alifestd_find_root_ids,
    alifestd_make_empty,
    alifestd_try_add_ancestor_list_col,
    argsort,
    give_len,
)
from ._Searchtable import Searchtable


def _collapse_indistinguishable_children(
    table: Searchtable, cur_node: int
) -> None:
    """Collapse that have been made collapsed-away precursors."""
    # group nodes made indistinguishable by collapsed precursors...
    groups = collections.defaultdict(list)
    for child in table.iter_inner_search_children_of(cur_node):
        key = (table.get_rank_of(child), table.get_differentia_of(child))
        groups[key].append(child)
    assert len(groups) == len(set(key[1] for key in groups.keys()))
    # ^^^ check assumption used in cpp implementation (differentiae are unique)
    for group in groups.values():
        winner, *losers = sorted(group)
        for loser in losers:  # keep only the 0th tiebreak winner
            # reassign loser's children to winner
            # must grab a copy of inner children to prevent
            # iterator invalidation
            for loser_child in [*table.iter_inner_search_children_of(loser)]:
                table.detach_search_parent(loser_child)
                table.attach_search_parent(
                    taxon_id=loser_child, parent_id=winner
                )
            # detach loser from search trie
            table.detach_search_parent(loser)


def _consolidate_if_rank_dropped(
    table: Searchtable, *, cur_node: int, next_rank: int
) -> None:
    """Collapse away a descendant nodes from "search trie" if their ranks have
    been dropped by stratum retention policy.

    Ranks are recognized as "dropped" if `next_rank` skips past them. Performs
    a depth-first search to discover all nodes with ranks prior to `next_rank`.
    """

    next_child = next(table.iter_inner_search_children_of(cur_node), None)  # type: ignore
    if next_child is None or table.get_rank_of(next_child) >= next_rank:
        return

    node_stack = [*table.iter_inner_search_children_of(cur_node)]

    # collapse away nodes with ranks that have been dropped
    while node_stack:
        pop_node = node_stack.pop()
        table.detach_search_parent(pop_node)
        # must grab a copy of inner children to prevent iterator
        # invalidation
        for grandchild in [*table.iter_inner_search_children_of(pop_node)]:
            # reattach dropped's children
            if table.get_rank_of(grandchild) >= next_rank:
                table.detach_search_parent(grandchild)
                table.attach_search_parent(
                    taxon_id=grandchild, parent_id=cur_node
                )
            else:
                node_stack.append(grandchild)

    _collapse_indistinguishable_children(table, cur_node)


def _place_allele(
    table: Searchtable,
    *,
    cur_node: int,
    next_rank: int,
    next_differentia: int,
) -> int:
    """Descends the subtrie from `cur_node` to retrieve the child node
    consistent with the given rank/differentia combination, creating a new
    node if necessary."""

    for child in table.iter_inner_search_children_of(cur_node):
        # check immediate children for next allele
        rank_matches = table.get_rank_of(child) == next_rank
        differentia_matches = (
            table.get_differentia_of(child) == next_differentia
        )
        if rank_matches and differentia_matches:
            return child
    else:
        # if no congruent node exists, create a new node
        return table.create_offspring(
            differentia=next_differentia,
            parent_id=cur_node,
            rank=next_rank,
        )


def _insert_artifact(
    table: "Searchtable",
    *,
    ranks: typing.Iterable[int],
    differentiae: typing.Iterable[int],
    taxon_label: int,
    num_strata_deposited: int,
) -> None:
    """Insert a taxon into the trie, ultimately resulting in the creation
    of an additional leaf node and --- if necessary --- a unifurcating chain of
    inner nodes subtending it."""

    cur_node = 0  # root
    for next_rank, next_differentia in zip(ranks, differentiae):
        _consolidate_if_rank_dropped(
            table=table, cur_node=cur_node, next_rank=next_rank
        )
        cur_node = _place_allele(
            table,
            cur_node=cur_node,
            next_rank=next_rank,
            next_differentia=next_differentia,
        )

    table.create_offspring(  # leaf node
        differentia=None,
        parent_id=cur_node,
        rank=num_strata_deposited - 1,
        taxon_label=taxon_label,
    )


def _finalize_records(
    records: typing.List[dict], *, force_common_ancestry: bool
) -> pd.DataFrame:
    """Collate as phylogeny dataframe in alife standard format."""
    df = pd.DataFrame(records)
    df["id"] = df["taxon_id"].astype(np.uint64)
    df["ancestor_id"] = df["ancestor_id"].astype(np.uint64)
    df["origin_time"] = df["rank"].astype(np.uint64)

    df = df[["id", "ancestor_id", "origin_time", "taxon_label"]]
    df = alifestd_try_add_ancestor_list_col(df, mutate=True)
    root_id = mit.one(alifestd_find_root_ids(df))
    multiple_true_roots = alifestd_count_children_of_asexual(df, root_id) > 1
    if multiple_true_roots and not force_common_ancestry:
        raise ValueError(
            "Reconstruction resulted in multiple independent trees, "
            "due to artifacts definitively sharing no common ancestor. "
            "Consider setting force_common_ancestry=True.",
        )
    df = alifestd_collapse_unifurcations(df, mutate=True)

    return df


def build_tree_searchtable_python(
    population: typing.Sequence[HereditaryStratigraphicArtifact],
    taxon_labels: typing.Optional[typing.Iterable] = None,
    progress_wrap: typing.Optional[typing.Callable] = None,
    force_common_ancestry: bool = False,
) -> pd.DataFrame:
    """Pure-python implementation of `build_tree_searchtable`.

    Parameters
    ----------
    population: Sequence[HereditaryStratigraphicArtifact]
        Hereditary stratigraphic columns corresponding to extant population members.

        Each member of population will correspond to a unique leaf node in the
        reconstructed tree.
    taxon_labels: Optional[Iterable], optional
        How should leaf nodes representing extant hereditary stratigraphic
        columns be named?

        Label order should correspond to the order of corresponding hereditary
        stratigraphic columns within `population`. If None, taxons will be
        named according to their numerical index.
    force_common_ancestry: bool, default False
        How should columns that definitively share no common ancestry be
        handled?

        If set to True, treat columns with no common ancestry as if they
        shared a common ancestor immediately before the genesis of the
        lineages. If set to False, columns within `population` that
        definitively do not share common ancestry will raise a ValueError.
    progress_wrap : Callable, optional
        Pass tqdm or equivalent to display a progress bar.

    Returns
    -------
    pd.DataFrame
        The reconstructed phylogenetic tree in alife standard format.

    Notes
    -----
    See `build_tree_searchtable` for algorithm overview.
    """
    pop_len = len(population)
    if pop_len == 0:
        return alifestd_make_empty()

    taxon_labels = [*opyt.or_value(taxon_labels, map(str, range(pop_len)))]
    sort_order = argsort([x.GetNumStrataDeposited() for x in population])
    sorted_population = [*map(population.__getitem__, sort_order)]
    taxon_labels = [*map(taxon_labels.__getitem__, sort_order)]

    table = Searchtable()
    for taxon_label, artifact in opyt.or_value(progress_wrap, lambda x: x)(
        give_len(zip(taxon_labels, sorted_population), len(sorted_population)),
    ):
        _insert_artifact(
            table,
            ranks=artifact.IterRetainedRanks(),
            differentiae=artifact.IterRetainedDifferentia(),
            taxon_label=taxon_label,
            num_strata_deposited=artifact.GetNumStrataDeposited(),
        )

    return _finalize_records(
        table.to_records(),
        force_common_ancestry=force_common_ancestry,
    )



import typing

from .._detail import PolicySpecBase


class PolicySpec(PolicySpecBase):
    """Contains all policy parameters, if any."""

    def __eq__(self: "PolicySpec", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    def __repr__(self: "PolicySpec") -> str:
        return f"""{
            self.GetAlgoIdentifier()
        }.{
            PolicySpec.__qualname__
        }()"""

    def __str__(self: "PolicySpec") -> str:
        return self.GetAlgoTitle()

    def GetEvalCtor(self: "PolicySpec") -> str:
        return f"hstrat.{self!r}"

    @staticmethod
    def GetAlgoIdentifier() -> str:
        """Get programatic name for underlying retention algorithm."""
        return __package__.split(".")[-1]

    @staticmethod
    def GetAlgoTitle() -> str:
        """Get human-readable name for underlying retention algorithm."""
        return "Nominal Resolution Stratum Retention Algorithm"



from ...._auxiliary_lib import jit


@jit(nopython=True)
def _get_niche_id(
    population_loc: int, island_niche_size: int, num_niches: int
) -> int:
    """Get population slot's niche ID."""
    # niches are contiguous and sequential within islands
    # if multiple islands are simulated, niches are fragmented in round robin
    # order across population
    return (population_loc // island_niche_size) % num_niches



import typing

from ...._detail import PolicyCouplerBase
from ...._impl import GenDropRanksFromPredKeepRank
from ..._PolicySpec import PolicySpec
from ..._impl import calc_provided_uncertainty


class _PredKeepRank:
    """Functor to implement the recency-proportional resolution stratum
    retention policy, for use with HereditaryStratigraphicColumn.

    This functor enacts the recency-proportional resolution policy by
    specifying the set of strata ranks that should be purged from a hereditary
    stratigraphic column when the nth stratum is deposited.
    """

    def __init__(
        self: "_PredKeepRank",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(self: "_PredKeepRank", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    # TODO refactor to not be recursive?
    @staticmethod
    def _do_call(
        policy: PolicyCouplerBase,
        num_stratum_depositions_completed: int,
        stratum_rank: int,
    ) -> bool:
        """Implementation for __call__ to faciliate external (but within-
        library) calls."""
        resolution = policy.GetSpec().GetRecencyProportionalResolution()

        # to satisfy requirements of HereditaryStratigraphicColumn impl
        # we must always keep root ancestor and newest stratum
        if stratum_rank in (0, num_stratum_depositions_completed):
            return True
        elif num_stratum_depositions_completed <= resolution:
            return True

        provided_uncertainty = calc_provided_uncertainty(
            resolution,
            num_stratum_depositions_completed,
        )
        # see CalcRankAtColumnIndex
        greatest_viable_rank = (
            num_stratum_depositions_completed
            - provided_uncertainty * (resolution + 1)
        )
        # see CalcRankAtColumnIndex
        num_interval_steps = (
            greatest_viable_rank + provided_uncertainty
        ) // provided_uncertainty
        cutoff_rank = num_interval_steps * provided_uncertainty

        # logically,  we could just test
        #   if stratum_rank == provided_uncertainty: return True
        # but we are guaranteed to eventually return True under the weaker
        # condition
        #   if stratum_rank % provided_uncertainty == 0
        # so as an optimization go ahead and return True now if it holds
        if stratum_rank % provided_uncertainty == 0:
            return True
        elif stratum_rank <= cutoff_rank:
            return False
        else:
            assert cutoff_rank
            return _PredKeepRank._do_call(
                policy,
                num_stratum_depositions_completed - cutoff_rank,
                stratum_rank - cutoff_rank,
            )

    def __call__(
        self: "_PredKeepRank",
        policy: PolicyCouplerBase,
        num_stratum_depositions_completed: int,
        stratum_rank: int,
    ) -> bool:
        """Decide if a stratum within the stratagraphic column should be
        retained or purged.

        Every time a new stratum is deposited, this method is called on each
        stratum present in a HereditaryStratigraphicColumn to determine whether
        it should be retained. Strata that return False are immediately purged
        from the column, meaning that for a stratum to persist it must earn a
        True result from this method each and every time a new stratum is
        deposited.

        Parameters
        ----------
        policy: Policy
            Policy this functor enacts.
        stratum_rank : int
            The number of strata that were deposited before the stratum under
            consideration for retention.
        num_stratum_depositions_completed : int
            The number of strata that have already been deposited, not
            including the latest stratum being deposited which prompted the
            current purge operation.

        Returns
        -------
        bool
            True if the stratum should be retained, False otherwise.

        See Also
        --------
        recency_proportional_resolution_algo:
            For details on the rationale, implementation, and guarantees of the
            recency-proportional resolution stratum retention policy.
        """
        return _PredKeepRank._do_call(
            policy,
            num_stratum_depositions_completed,
            stratum_rank,
        )


FromPredKeepRank = GenDropRanksFromPredKeepRank(_PredKeepRank)



import typing

import pandas as pd

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_is_topologically_sorted import alifestd_is_topologically_sorted
from ._alifestd_parse_ancestor_ids import alifestd_parse_ancestor_ids
from ._alifestd_topological_sort import alifestd_topological_sort
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col


def alifestd_mark_root_id(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
    selector: typing.Callable = min,
) -> pd.DataFrame:
    """Add column `root_id`, containing the `id` of entries' ultimate ancestor.

    For sexual data, the field `root_id` is chosen according to the selection
    of callable `selector` over parents' `root_id` values. Note that subsets
    within a connected component may be marked with different `root_id` values.
    To create a component id that is consistent within connected components,
    a backward pass could be performed that updates ancestors' values if they
    are greater than that of each descendant.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if not alifestd_is_topologically_sorted(phylogeny_df):
        phylogeny_df = alifestd_topological_sort(phylogeny_df, mutate=True)

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df.reset_index(drop=True, inplace=True)
    else:
        phylogeny_df.index = phylogeny_df["id"]

    phylogeny_df["root_id"] = phylogeny_df["id"]
    if "ancestor_id" in phylogeny_df.columns:  # asexual
        root_id_col = phylogeny_df["root_id"]
        ancestor_id_col = phylogeny_df["ancestor_id"]
        for index in phylogeny_df.index:
            ancestor_id = ancestor_id_col.at[index]
            root_id_col.at[index] = root_id_col.at[ancestor_id]
    else:  # sexual
        root_id_col = phylogeny_df["root_id"]
        ancestor_list_col = phylogeny_df["ancestor_list"]
        for index in phylogeny_df.index:
            ancestor_list = ancestor_list_col.at[index]
            ancestor_ids = alifestd_parse_ancestor_ids(ancestor_list)
            candidate_roots = [*map(root_id_col.at.__getitem__, ancestor_ids)]
            # "or" covers genesis empty list case
            root_id_col.at[index] = selector(candidate_roots or [index])

    return phylogeny_df



"""Tools to create extant populations of hereditary stratigraph columns
as if they had resulted from a given template phylogeny."""

from ..._auxiliary_lib import lazy_attach_stub

__getattr__, __dir__, __all__ = lazy_attach_stub(__name__, __file__)
del lazy_attach_stub



import itertools as it

import numpy as np
import pytest

from hstrat._auxiliary_lib import apply_swaps


@pytest.mark.parametrize(
    "pop_size",
    [
        0,
        1,
        2,
        5,
        100,
    ],
)
def test_apply_zero_swaps(pop_size):

    iota = np.arange(pop_size)
    apply_swaps(iota, np.array([], dtype=int), np.array([], dtype=int))
    assert len(set(iota)) == len(iota)
    assert all(iota == np.arange(pop_size))


@pytest.mark.parametrize(
    "pop_size",
    [
        1,
        2,
        5,
        100,
    ],
)
def test_apply_one_swap(pop_size):

    for operand in [np.array([0]), np.array([-1])]:
        iota = np.arange(pop_size)
        apply_swaps(iota, operand, operand)
        assert all(iota == np.arange(pop_size))

    for operands in it.permutations([np.array([0]), np.array([-1])]):
        iota = np.arange(pop_size)
        apply_swaps(iota, *operands)
        assert len(set(iota)) == len(iota)
        assert iota[0] == pop_size - 1
        assert iota[-1] == 0

    target_idx = min(pop_size - 1, 3)
    for operands in it.permutations(
        [np.array([target_idx]), np.array([target_idx - 1])]
    ):
        iota = np.arange(pop_size)

        apply_swaps(iota, *operands)
        assert len(set(iota)) == len(iota)
        assert iota[target_idx] == max(target_idx - 1, 0)
        assert iota[target_idx - 1] == target_idx


@pytest.mark.parametrize(
    "pop_size",
    [
        3,
        5,
        100,
    ],
)
def test_apply_two_swaps(pop_size):

    iota = np.arange(pop_size)
    apply_swaps(iota, np.array([0, 1]), np.array([1, 2]))
    assert iota[0] == 1
    assert iota[1] == 2
    assert iota[2] == 0

    if pop_size > 3:
        iota = np.arange(pop_size)
        apply_swaps(iota, np.array([0, -1]), np.array([1, -2]))
        assert iota[0] == 1
        assert iota[1] == 0
        assert iota[-1] == pop_size - 2
        assert iota[-2] == pop_size - 1


@pytest.mark.parametrize(
    "pop_size",
    [
        1,
        2,
        5,
        100,
    ],
)
@pytest.mark.parametrize(
    "num_swaps",
    [
        0,
        1,
        2,
        5,
        100,
        1000,
    ],
)
def test_apply_swaps_conserves(pop_size, num_swaps):
    iota = np.arange(pop_size)
    swap_from_idxs = np.random.randint(pop_size, size=num_swaps)
    swap_to_idxs = np.random.randint(pop_size, size=num_swaps)
    apply_swaps(iota, swap_from_idxs, swap_to_idxs)
    assert len(set(iota)) == len(iota)


@pytest.mark.parametrize(
    "pop_size",
    [
        1,
        2,
        5,
        100,
    ],
)
@pytest.mark.parametrize(
    "num_swaps",
    [
        0,
        1,
        2,
        5,
        100,
        1000,
    ],
)
def test_apply_swaps_decomposes(pop_size, num_swaps):
    for split in range(pop_size):
        iota = np.arange(pop_size)
        swap_from_idxs = np.random.randint(pop_size, size=num_swaps)
        swap_to_idxs = np.random.randint(pop_size, size=num_swaps)
        apply_swaps(iota, swap_from_idxs[:split], swap_to_idxs[:split])
        apply_swaps(iota, swap_from_idxs[split:], swap_to_idxs[split:])

        iota2 = np.arange(pop_size)
        apply_swaps(iota2, swap_from_idxs, swap_to_idxs)

        assert all(iota == iota2)


@pytest.mark.parametrize(
    "pop_size",
    [
        1,
        2,
        5,
        100,
    ],
)
@pytest.mark.parametrize(
    "num_swaps",
    [
        0,
        1,
        2,
        5,
        100,
        1000,
    ],
)
def test_apply_swaps_reversible(pop_size, num_swaps):
    iota = np.arange(pop_size)
    swap_from_idxs = np.random.randint(pop_size, size=num_swaps)
    swap_to_idxs = np.random.randint(pop_size, size=num_swaps)
    apply_swaps(iota, swap_from_idxs, swap_to_idxs)
    apply_swaps(iota, swap_from_idxs[::-1], swap_to_idxs[::-1])
    assert all(iota == np.arange(pop_size))



import copy
import itertools as it

import pytest

from hstrat import hstrat
from hstrat._auxiliary_lib import (
    is_strictly_increasing,
    random_choice_generator,
)


@pytest.mark.parametrize(
    "differentia_bit_width", [1, 2, 7, 8, 16, 32, 64, 129]
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2
        ),
    ],
)
@pytest.mark.parametrize("synchronous_generations", [True, False])
@pytest.mark.parametrize("population_size", [1, 2, 9])
@pytest.mark.parametrize("num_generations", [0, 1, 257])
def test_init_and_getters(
    differentia_bit_width,
    retention_policy,
    synchronous_generations,
    population_size,
    num_generations,
):
    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_bit_width,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(population_size)
    ]
    reproduction_queue = (
        it.cycle if synchronous_generations else random_choice_generator
    )(population)
    for selection in it.islice(
        reproduction_queue, 0, num_generations * population_size
    ):
        selection.DepositStratum()

    assemblage = hstrat.pop_to_assemblage(population)

    for specimen, column in zip(assemblage.BuildSpecimens(), population):
        assert (
            specimen.GetNumStrataDeposited() == column.GetNumStrataDeposited()
        )
        assert (
            specimen.GetStratumDifferentiaBitWidth()
            == column.GetStratumDifferentiaBitWidth()
        )
        assert specimen.GetNumStrataRetained() == column.GetNumStrataRetained()
        assert {*specimen.GetDifferentiaVals()} >= {
            *column.IterRetainedDifferentia()
        }
        assert [
            *specimen.GetDifferentiaVals()[~specimen.GetStratumMask()]
        ] >= [*column.IterRetainedDifferentia()]
        assert {*specimen.GetRankIndex()} >= {*column.IterRetainedRanks()}
        assert [*specimen.GetRankIndex()[~specimen.GetStratumMask()]] == [
            *column.IterRetainedRanks()
        ]
        assert is_strictly_increasing(specimen.GetRankIndex())

        assert [*specimen.GetData().dropna()] == [
            *column.IterRetainedDifferentia()
        ]
        assert {*specimen.GetData().index} >= {*column.IterRetainedRanks()}
        assert is_strictly_increasing(specimen.GetData().index)

        for index in range(column.GetNumStrataRetained()):
            assert specimen.GetRankAtColumnIndex(
                index
            ) == column.GetRankAtColumnIndex(index)

        assert [*specimen.IterRetainedRanks()] == [*column.IterRetainedRanks()]
        assert [*specimen.IterRetainedDifferentia()] == [
            *column.IterRetainedDifferentia()
        ]

        assert (
            specimen.GetNumDiscardedStrata() == column.GetNumDiscardedStrata()
        )
        assert specimen.HasDiscardedStrata() == column.HasDiscardedStrata()


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2
        ),
    ],
)
def test_IterRankDifferentiaZip(retention_policy):
    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
    )
    c2 = c1.CloneNthDescendant(10)
    c3 = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
    ).CloneNthDescendant(4)

    for __ in range(100):
        c1.DepositStratum()
        c2.DepositStratum()
        c3.DepositStratum()
        for subpopulation in (
            [c1, c2],
            [c1, c2, c3],
        ):
            assemblage = hstrat.pop_to_assemblage(subpopulation)
            for specimen in assemblage.BuildSpecimens():
                assert [*specimen.IterRankDifferentiaZip()] == [
                    *zip(
                        specimen.IterRetainedRanks(),
                        specimen.IterRetainedDifferentia(),
                    )
                ]
                iter_ = specimen.IterRankDifferentiaZip(copyable=True)
                iter_copy = copy.copy(iter_)
                next(iter_copy)
                assert [*iter_copy] == [
                    *zip(
                        specimen.IterRetainedRanks(),
                        specimen.IterRetainedDifferentia(),
                    )
                ][1:]
                assert [*iter_] == [
                    *zip(
                        specimen.IterRetainedRanks(),
                        specimen.IterRetainedDifferentia(),
                    )
                ]



import typing

from ..._detail import PolicyCouplerBase
from .._PolicySpec import PolicySpec


class CalcNumStrataRetainedExact:
    """Functor to provide member function implementation in Policy class."""

    def __init__(
        self: "CalcNumStrataRetainedExact",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(self: "CalcNumStrataRetainedExact", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcNumStrataRetainedExact",
        policy: PolicyCouplerBase,
        num_strata_deposited: int,
    ) -> int:
        """Exactly how many strata are retained after n deposited?"""
        spec = policy.GetSpec()

        if num_strata_deposited == 0:
            return 0

        uncertainty = spec.GetFixedResolution()
        newest_stratum_rank = num_strata_deposited - 1
        # +1 for 0'th rank stratum
        num_strata_at_uncertainty_intervals = (
            newest_stratum_rank // uncertainty + 1
        )
        newest_stratum_distinct_from_uncertainty_intervals = (
            newest_stratum_rank % uncertainty != 0
        )
        return (
            num_strata_at_uncertainty_intervals
            + newest_stratum_distinct_from_uncertainty_intervals
        )



import typing

import numpy as np

from ._jit import jit


@jit(nopython=True)
def iter_monotonic_equivalencies(
    first: np.ndarray,
    second: np.ndarray,
    start: typing.Tuple[int, int] = (0, 0),
) -> typing.Iterator[typing.Tuple[int, int]]:
    """Find the indices of equivalent elements in two sorted arrays.

    Given two sorted arrays `first` and `second`, this function returns an
    iterator that yields pairs of indices (i, j) such that first[i] ==
    second[j]. The iteration stops when one of the arrays is fully traversed.

    Parameters
    ----------
    first : np.ndarray
        A sorted numpy array.
    second : np.ndarray
        A sorted numpy array.
    start : tuple of int, default (0, 0)
        A pair of indices (i, j) where the search should start. The
        search will begin at the element first[i] and second[j]. Default
        is (0, 0).

    Yields
    ------
    (i, j) : tuple of int
        A tuple of indices (i, j) where first[i] == second[j].

    Examples
    --------
    >>> first = np.array([1, 2, 3, 4, 5])
    >>> second = np.array([2, 4, 6, 8, 10])
    >>> for i, j in iter_monotonic_equivalencies(first, second, start=(1, 0)):
    ...     print(f"{i} {j}")
    1 0
    3 1

    """
    idx1, idx2 = start
    assert 0 <= idx1
    assert 0 <= idx2

    while idx1 < len(first) and idx2 < len(second):
        val1 = first[idx1]
        val2 = second[idx2]

        if val1 < val2:
            idx1 += 1
        elif val1 > val2:
            idx2 += 1
        else:  # val1 == val2
            yield (idx1, idx2)
            idx1 += 1
            idx2 += 1



import typing

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ._calc_rank_of_mrca_bounds_provided_confidence_level import (
    calc_rank_of_mrca_bounds_provided_confidence_level,
)


def calc_ranks_since_mrca_bounds_provided_confidence_level(
    focal: HereditaryStratigraphicArtifact,
    other: HereditaryStratigraphicArtifact,
    prior: typing.Literal["arbitrary"],
    requested_confidence_level: float = 0.95,
) -> float:
    """Calculate provided confidence for a MRCA generation estimate.

    With what actual confidence is the true rank of the MRCA captured within the calculated estimate bounds for a requested confidence level?
    Guaranteed greater than or equal to the requested confidence level.

    The same argument may be provided for focal and other.
    """
    return calc_rank_of_mrca_bounds_provided_confidence_level(
        focal,
        other,
        prior,
        requested_confidence_level,
    )



import typing
from unittest import mock

import more_itertools as mit
import numpy as np
import opytional as opyt
import pandas as pd
import polars as pl

from ...._auxiliary_lib import (
    HereditaryStratigraphicArtifact,
    alifestd_make_empty,
    alifestd_try_add_ancestor_list_col,
    argsort,
)
from ._build_tree_searchtable_cpp_impl_stub import (
    Records,
    build_tree_searchtable_cpp_from_exploded,
    build_tree_searchtable_cpp_from_nested,
    collapse_dropped_unifurcations,
    extend_tree_searchtable_cpp_from_exploded,
    records_to_dict,
)


def _finalize_records(
    records: dict[str, np.ndarray],
    sorted_labels: typing.List[str],
    force_common_ancestry: bool,
) -> pd.DataFrame:
    """Collate as phylogeny dataframe in alife standard format."""
    df = pd.DataFrame(records)
    df["origin_time"] = df["rank"]
    df["taxon_label"] = [
        str(sorted_labels[i]) if i != (2**64 - 1) else "_inner_node"
        for i in df["dstream_data_id"]
    ]

    multiple_true_roots = (
        (df["id"] != 0) & (df["ancestor_id"] == 0)
    ).sum() > 1
    if multiple_true_roots and not force_common_ancestry:
        raise ValueError(
            "Reconstruction resulted in multiple independent trees, "
            "due to artifacts definitively sharing no common ancestor. "
            "Consider setting force_common_ancestry=True.",
        )

    return alifestd_try_add_ancestor_list_col(df, mutate=True)


def _explode_population(
    sorted_population: typing.Sequence[HereditaryStratigraphicArtifact],
) -> pl.DataFrame:
    """Create DataFrame with one row for each retained stratum across all
    population members."""
    return pl.DataFrame(
        {
            "data_ids": [
                i
                for i, ann in enumerate(sorted_population)
                for __ in range(ann.GetNumStrataRetained())
            ],
            "num_strata_depositeds": [
                ann.GetNumStrataDeposited()
                for ann in sorted_population
                for __ in range(ann.GetNumStrataRetained())
            ],
            "ranks": [
                rank
                for ann in sorted_population
                for rank in ann.IterRetainedRanks()
            ],
            "differentiae": [
                differentia
                for ann in sorted_population
                for differentia in ann.IterRetainedDifferentia()
            ],
        },
        schema={
            "data_ids": pl.UInt64,
            "num_strata_depositeds": pl.UInt64,
            "ranks": pl.UInt64,
            "differentiae": pl.UInt64,
        },
    )


def build_tree_searchtable_cpp(
    population: typing.Sequence[HereditaryStratigraphicArtifact],
    taxon_labels: typing.Optional[typing.Iterable] = None,
    progress_wrap: typing.Optional[typing.Callable] = None,
    force_common_ancestry: bool = False,
    _entry_point: typing.Literal[
        "batched_small",
        "batched_medium",
        "batched_large",
        "nested",
        "exploded",
    ] = "nested",
) -> pd.DataFrame:
    """C++-based implementation of `build_tree_searchtable`.

    Parameters
    ----------
    population: Sequence[HereditaryStratigraphicArtifact]
        Hereditary stratigraphic columns corresponding to extant population members.

        Each member of population will correspond to a unique leaf node in the
        reconstructed tree.
    taxon_labels: Optional[Iterable], optional
        How should leaf nodes representing extant hereditary stratigraphic
        columns be named?

        Label order should correspond to the order of corresponding hereditary
        stratigraphic columns within `population`. If None, taxons will be
        named according to their numerical index.
    force_common_ancestry: bool, default False
        How should columns that definitively share no common ancestry be
        handled?

        If set to True, treat columns with no common ancestry as if they
        shared a common ancestor immediately before the genesis of the
        lineages. If set to False, columns within `population` that
        definitively do not share common ancestry will raise a ValueError.
    progress_wrap : Callable, optional
        Pass tqdm or equivalent to display a progress bar.
    _entry_point : Literal["nested", "exploded"], default "nested"
        Which implementation interface should be called?

        For internal use in testing.

    Returns
    -------
    pd.DataFrame
        The reconstructed phylogenetic tree in alife standard format.

    Notes
    -----
    See `build_tree_searchtable` for algorithm overview.
    """
    pop_len = len(population)
    if pop_len == 0:
        return alifestd_make_empty()

    taxon_labels = list(
        opyt.or_value(
            taxon_labels,
            map(int, range(pop_len)),
        )
    )
    sort_order = argsort([x.GetNumStrataDeposited() for x in population])
    sorted_labels = [taxon_labels[i] for i in sort_order]
    sorted_population = [population[i] for i in sort_order]

    if _entry_point == "nested":
        records = build_tree_searchtable_cpp_from_nested(
            [*range(len(sorted_population))],
            [ann.GetNumStrataDeposited() for ann in sorted_population],
            [[*ann.IterRetainedRanks()] for ann in sorted_population],
            [[*ann.IterRetainedDifferentia()] for ann in sorted_population],
            opyt.or_value(progress_wrap, mock.Mock()),
        )
    elif _entry_point == "exploded":
        exploded_df = _explode_population(sorted_population)
        records = build_tree_searchtable_cpp_from_exploded(
            exploded_df["data_ids"].to_numpy(),
            exploded_df["num_strata_depositeds"].to_numpy(),
            exploded_df["ranks"].to_numpy(),
            exploded_df["differentiae"].to_numpy(),
            opyt.or_value(progress_wrap, mock.Mock()),
        )
    elif _entry_point.startswith("batched_"):
        exploded_df = _explode_population(sorted_population)
        batch_size = {
            "batched_small": 10,
            "batched_medium": 1_000,
            "batched_large": 10_000_000,
        }[_entry_point]
        records = Records(len(exploded_df) * 4)
        for partition_dfs in mit.sliced(
            exploded_df.partition_by("data_ids"), batch_size
        ):
            slice_df = pl.concat(partition_dfs)
            extend_tree_searchtable_cpp_from_exploded(
                records,
                slice_df["data_ids"].to_numpy(),
                slice_df["num_strata_depositeds"].to_numpy(),
                slice_df["ranks"].to_numpy(),
                slice_df["differentiae"].to_numpy(),
                opyt.or_value(progress_wrap, mock.Mock()),
            )
            records = collapse_dropped_unifurcations(records)
        records = records_to_dict(records)
    else:
        raise ValueError(f"Invalid entry point: {_entry_point}")

    assert len(records["dstream_data_id"]) >= len(sorted_population) + 1
    return _finalize_records(
        records,
        sorted_labels,
        force_common_ancestry,
    )



import pandas as pd

from ._alifestd_count_leaf_nodes import alifestd_count_leaf_nodes
from ._alifestd_count_root_nodes import alifestd_count_root_nodes
from ._alifestd_count_unifurcations import alifestd_count_unifurcations


def alifestd_calc_polytomic_index(phylogeny_df: pd.DataFrame) -> int:
    """Count how many fewer inner nodes are contained in phylogeny than expected
    if strictly bifurcationg.

    Excludes unifurcations from calculation.
    """
    num_leaf_nodes = alifestd_count_leaf_nodes(phylogeny_df)
    num_root_nodes = alifestd_count_root_nodes(phylogeny_df)
    expected_rows_if_bifurcating = max(2 * num_leaf_nodes - num_root_nodes, 0)
    num_unifurcations = alifestd_count_unifurcations(phylogeny_df)
    num_non_unifurcating_rows = len(phylogeny_df) - num_unifurcations
    res = expected_rows_if_bifurcating - num_non_unifurcating_rows
    assert 0 <= res < max(expected_rows_if_bifurcating, 1)
    return res



import typing

import pandas as pd

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ._impl import build_tree_biopython_distance


def build_tree_nj(
    population: typing.Sequence[HereditaryStratigraphicArtifact],
    estimator: str,
    prior: typing.Union[str, typing.Any],
    taxon_labels: typing.Optional[typing.Iterable] = None,
    force_common_ancestry: bool = False,
    negative_origin_time_correction_method: typing.Optional[str] = None,
) -> pd.DataFrame:
    """Estimate the phylogenetic history among hereditary stratigraphic
    columns by using the "unweighted pair group method with neighbor joinint
    distance-based reconstruction method.

    This phylogenetic reconstruction approach is generally unfavorable, incuring
    `O(n^2)` runtime complexity and providing reconstructions that occasionally
    conflict with the hereditary stratigraphic record.

    Parameters
    ----------
    population: Sequence[HereditaryStratigraphicArtifact]
        Hereditary stratigraphic columns corresponding to extant population
        members.

        Each member of population will correspond to a unique leaf node in the
        reconstructed tree.
    estimator : {"maximum_likelihood", "unbiased"}
        What estimation method should be used? Options are "maximum_likelihood"
        or "unbiased".

        See `estimate_ranks_since_mrca_with` for discussion of estimator
        options.
    prior : {"arbitrary", "uniform"} or object implementing prior interface
        Prior probability density distribution over possible generations of the
        MRCA.

        See `estimate_rank_of_mrca_between` for discussion of prior
        options.
    taxon_labels: Optional[Iterable], optional
        How should leaf nodes representing extant hereditary stratigraphic
        columns be named?

        Label order should correspond to the order of corresponding hereditary
        stratigraphic columns within `population`. If None, taxons will be
        named according to their numerical index.
    force_common_ancestry: bool, default False
        How should columns that definitively share no common ancestry be handled?

        If set to True, treat columns with no common ancestry as if they
        shared a common ancestor immediately before the genesis of the
        lineages. If set to False, columns within `population` that
        definitively do not share common ancestry will raise a ValueError.
    negative_origin_time_correction_method
        : {"truncate", "shift", "rescale"}, optional
        How should negative origin time estimates be corrected?

    Returns
    -------
    pd.DataFrame
        The reconstructed phylogenetic tree in alife standard format.
    """
    return build_tree_biopython_distance(
        "nj",
        population,
        estimator,
        prior,
        taxon_labels,
        force_common_ancestry,
        negative_origin_time_correction_method,
    )



import pandas as pd

from hstrat._auxiliary_lib import (
    alifestd_count_leaf_nodes,
    alifestd_make_empty,
)


def test_empty_df():
    assert alifestd_count_leaf_nodes(alifestd_make_empty()) == 0


def test_singleton_df():
    df = pd.DataFrame(
        {
            "id": [0],
            "ancestor_list": [[None]],
        }
    )
    assert alifestd_count_leaf_nodes(df) == 1


def test_sexual_df1():
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4],
            "ancestor_list": ["[None]", "[0]", "[0]", "[1, 0]", "[1]"],
        }
    )
    assert alifestd_count_leaf_nodes(df) == 3


def test_sexual_df2():
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4],
            "ancestor_list": [[None], [0], [0], [1, 0], [1]],
        }
    )
    assert alifestd_count_leaf_nodes(df) == 3


def test_polytomy_df():
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4],
            "ancestor_list": [[None], [0], [0], [0], [1]],
        }
    )
    assert alifestd_count_leaf_nodes(df) == 3


def test_multiple_trees_df1():
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "ancestor_list": [[None], [None], [0], [2], [2], [3]],
        }
    )
    assert alifestd_count_leaf_nodes(df) == 3


def test_multiple_trees_df2():
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "ancestor_list": [[None], [None], [0], [1], [2], [3]],
        }
    )
    assert alifestd_count_leaf_nodes(df) == 2


def test_strictly_bifurcating_df():
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4],
            "ancestor_list": [[None], [0], [0], [1], [1]],
        }
    )
    assert alifestd_count_leaf_nodes(df) == 3



import pandas as pd


def swap_rows_and_indices(df: pd.DataFrame, loc1, loc2) -> pd.DataFrame:
    """Swap positions of indices `loc1` and `loc2` and their corresponding
    rows."""
    indices = [*df.index]
    iloc1 = df.index.get_loc(loc1)
    iloc2 = df.index.get_loc(loc2)
    indices[iloc1], indices[iloc2] = indices[iloc2], indices[iloc1]
    return df.loc[indices]



import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_make_ancestor_list_col,
    alifestd_make_empty,
    alifestd_sample_triplet_comparisons_asexual,
)


def test_empty():
    with pytest.raises(ValueError):
        alifestd_sample_triplet_comparisons_asexual(
            alifestd_make_empty(), alifestd_make_empty(), "id"
        )


def test_undersize():
    df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "taxon_label": [4, 5, 6],
            "ancestor_id": [0, 0, 0],
        },
    )
    df["ancestor_list"] = alifestd_make_ancestor_list_col(
        df["id"], df["ancestor_id"]
    )
    for xdf in df[:1], df[:2], df:
        with pytest.raises(ValueError):
            alifestd_sample_triplet_comparisons_asexual(
                xdf, xdf.sample(frac=1), "taxon_label"
            )


@pytest.mark.parametrize(
    "df",
    [
        pd.DataFrame(
            {
                "id": [0, 1, 2, 3],
                "taxon_label": [0, 1, 2, 3],
                "ancestor_id": [0, 0, 0, 0],
            },
        ),
        pd.DataFrame(
            {
                "id": [0, 1, 2, 3, 4, 5],
                "taxon_label": ["0", "1", "2", "3", "4", "5"],
                "ancestor_id": [0, 0, 0, 0, 1, 2],
            },
        ),
        pd.DataFrame(
            {
                "id": [9, 1, 2, 3, 4, 5],
                "taxon_label": ["9", "1", "2", "3", "4", "5"],
                "ancestor_id": [9, 9, 9, 9, 1, 2],
            },
        ),
        pd.DataFrame(
            {
                "id": [0, 1, 2, 3, 4, 5, 6],
                "taxon_label": ["0", "1", "2", "3", "4", "5", "6"],
                "ancestor_id": [0, 0, 0, 0, 1, 2, 3],
            },
        ),
    ],
)
def test_polytomy_identical(df: pd.DataFrame):
    df["ancestor_list"] = alifestd_make_ancestor_list_col(
        df["id"], df["ancestor_id"]
    )
    for adf, bdf in (df, df), (df, df.sample(frac=1)):
        info_df = alifestd_sample_triplet_comparisons_asexual(
            adf, bdf, "taxon_label"
        )
        assert info_df["triplet match, lax"].all()
        assert info_df["triplet match, lax/strict"].all()
        assert info_df["triplet match, strict"].all()
        assert info_df["triplet match, strict/lax"].all()
        assert set(df.columns) < set(info_df.columns)


def test_differing_wrong1():
    adf = pd.DataFrame(
        {
            "id": reversed([9, 1, 2, 3, 4, 5]),
            "taxon_label": reversed(["0", "1", "2", "3", "4", "5"]),
            "ancestor_id": reversed([9, 9, 1, 2, 2, 1]),
        },
    )
    adf["ancestor_list"] = alifestd_make_ancestor_list_col(
        adf["id"], adf["ancestor_id"]
    )
    bdf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 5, 4],
            "taxon_label": ["0", "1", "2", "3", "5", "4"],
            "ancestor_id": [0, 0, 1, 2, 2, 1],
        },
    )
    bdf["ancestor_list"] = alifestd_make_ancestor_list_col(
        bdf["id"], bdf["ancestor_id"]
    )
    info_df = alifestd_sample_triplet_comparisons_asexual(
        adf, bdf, "taxon_label", n=200
    )
    assert len(info_df) == 200
    assert not info_df["triplet match, lax"].all()
    assert not info_df["triplet match, lax/strict"].all()
    assert not info_df["triplet match, strict"].all()
    assert not info_df["triplet match, strict/lax"].all()
    assert set(adf.columns) < set(info_df.columns)


def test_differing_polytomy():
    adf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "taxon_label": ["0", "1", "2", "3", "4", "5"],
            "ancestor_id": [0, 0, 1, 2, 2, 1],
        },
    )
    adf["ancestor_list"] = alifestd_make_ancestor_list_col(
        adf["id"], adf["ancestor_id"]
    )
    bdf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "taxon_label": ["0", "1", "2", "3", "4", "5"],
            "ancestor_id": [0, 0, 1, 2, 2, 2],
        },
    )
    bdf["ancestor_list"] = alifestd_make_ancestor_list_col(
        bdf["id"], bdf["ancestor_id"]
    )
    info_df = alifestd_sample_triplet_comparisons_asexual(
        adf, bdf, "id", n=200
    )
    assert info_df["triplet match, lax"].all()
    assert not info_df["triplet match, lax/strict"].all()
    assert not info_df["triplet match, strict"].all()
    assert info_df["triplet match, strict/lax"].all()
    assert set(adf.columns) < set(info_df.columns)



import numpy as np
import pandas as pd


def alifestd_find_root_ids(phylogeny_df: pd.DataFrame) -> np.ndarray:  # int
    """What ids have an empty `ancestor_list`?

    Input dataframe is not mutated by this operation.
    """

    root_df = phylogeny_df[
        phylogeny_df["ancestor_list"]
        .astype(str)
        .str.lower()
        .isin(("[none]", "[]"))
    ]
    return root_df["id"].to_numpy().copy()



import operator
import typing

from ....._auxiliary_lib import find_bounds
from ..._detail import PolicyCouplerBase
from .._PolicySpec import PolicySpec
from .._impl import get_retained_ranks


class CalcMrcaUncertaintyAbsExact:
    """Functor to provide member function implementation in Policy class."""

    def __init__(
        self: "CalcMrcaUncertaintyAbsExact",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(self: "CalcMrcaUncertaintyAbsExact", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcMrcaUncertaintyAbsExact",
        policy: PolicyCouplerBase,
        first_num_strata_deposited: int,
        second_num_strata_deposited: int,
        actual_rank_of_mrca: int,
    ) -> int:
        """Exactly how much uncertainty to estimate rank of MRCA?"""
        # rectify negative-indexed actual_rank_of_mrca
        if actual_rank_of_mrca is not None and actual_rank_of_mrca < 0:
            least_last_rank = min(
                first_num_strata_deposited - 1,
                second_num_strata_deposited - 1,
            )
            actual_rank_of_mrca += least_last_rank
            assert actual_rank_of_mrca >= 0

        least_num_strata_deposited = min(
            first_num_strata_deposited,
            second_num_strata_deposited,
        )
        least_last_rank = least_num_strata_deposited - 1

        retained_ranks = get_retained_ranks(
            policy,
            least_num_strata_deposited,
        )
        lower_bound, upper_bound = find_bounds(
            query=actual_rank_of_mrca,
            iterable=retained_ranks,
            filter_below=operator.le,
            filter_above=operator.gt,
            initializer=(0, least_num_strata_deposited),
        )

        assert lower_bound < upper_bound
        return upper_bound - lower_bound - 1



from . import trie_postprocess
from ._build_tree import build_tree
from ._build_tree_nj import build_tree_nj
from ._build_tree_searchtable import build_tree_searchtable
from ._build_tree_trie import build_tree_trie
from ._build_tree_trie_ensemble import build_tree_trie_ensemble
from ._build_tree_upgma import build_tree_upgma
from .trie_postprocess import (
    AssignDestructionTimeYoungestPlusOneTriePostprocessor,
    AssignOriginTimeExpectedValueTriePostprocessor,
    AssignOriginTimeNaiveTriePostprocessor,
    AssignOriginTimeNodeRankTriePostprocessor,
    AssignOriginTimeSampleNaiveTriePostprocessor,
    CompoundTriePostprocessor,
    PeelBackConjoinedLeavesTriePostprocessor,
    SampleAncestralRollbacksTriePostprocessor,
)

__all__ = [
    "build_tree_nj",
    "build_tree_searchtable",
    "build_tree_trie",
    "build_tree_trie_ensemble",
    "build_tree_upgma",
    "build_tree",
    "trie_postprocess",
    "AssignDestructionTimeYoungestPlusOneTriePostprocessor",
    "AssignOriginTimeExpectedValueTriePostprocessor",
    "AssignOriginTimeNaiveTriePostprocessor",
    "AssignOriginTimeNodeRankTriePostprocessor",
    "AssignOriginTimeSampleNaiveTriePostprocessor",
    "CompoundTriePostprocessor",
    "PeelBackConjoinedLeavesTriePostprocessor",
    "SampleAncestralRollbacksTriePostprocessor",
]



from hstrat._auxiliary_lib import alifestd_make_empty, alifestd_validate


def test_alifestd_make_empty():
    assert len(alifestd_make_empty()) == 0
    assert alifestd_validate(alifestd_make_empty())
    assert "id" in alifestd_make_empty()
    assert "ancestor_list" in alifestd_make_empty()


def test_alifestd_make_empty_ancestor_id():
    assert len(alifestd_make_empty(ancestor_id=True)) == 0
    assert alifestd_validate(alifestd_make_empty(ancestor_id=True))
    assert "id" in alifestd_make_empty(ancestor_id=True)
    assert "ancestor_list" in alifestd_make_empty(ancestor_id=True)
    assert "ancestor_id" in alifestd_make_empty(ancestor_id=True)



from ._HereditaryStratigraphicAssemblage import (
    HereditaryStratigraphicAssemblage,
)
from ._HereditaryStratigraphicAssemblageSpecimen import (
    HereditaryStratigraphicAssemblageSpecimen,
)
from ._HereditaryStratigraphicSpecimen import HereditaryStratigraphicSpecimen

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "HereditaryStratigraphicAssemblage",
    "HereditaryStratigraphicAssemblageSpecimen",
    "HereditaryStratigraphicSpecimen",
]



import itertools as it

import pytest

from hstrat._auxiliary_lib import estimate_binomial_p


def always_success_sampler():
    """A sampler that always returns success."""
    return True


def never_success_sampler():
    """A sampler that never returns success."""
    return False


_half_success_state = iter(it.cycle([True, False]))


def half_success_sampler():
    """A sampler that returns success half of the time."""
    # Alternate between True and False
    return next(_half_success_state)


@pytest.mark.parametrize(
    "sampler,expected_success",
    [
        (always_success_sampler, 1.0),
        (never_success_sampler, 0.0),
        (half_success_sampler, 0.5),
    ],
)
@pytest.mark.parametrize("precision", [0.02, 0.1])
def test_estimate_binomial_p(sampler, expected_success, precision):
    p_estimate, confint, nobs1 = estimate_binomial_p(
        sampler=sampler, precision=precision, confidence=0.8
    )
    assert abs(p_estimate - expected_success) <= precision
    assert 0 <= confint[1] - confint[0] <= precision
    assert nobs1 > 0

    p_estimate, confint, nobs2 = estimate_binomial_p(
        sampler=sampler, precision=precision, confidence=0.9
    )
    assert abs(p_estimate - expected_success) <= precision
    assert 0 <= confint[1] - confint[0] <= precision
    assert nobs2 > 0
    assert nobs2 > nobs1

    p_estimate, confint, nobs3 = estimate_binomial_p(
        sampler=sampler, precision=precision / 2, confidence=0.8
    )
    assert abs(p_estimate - expected_success) <= precision
    assert 0 <= confint[1] - confint[0] <= precision
    assert nobs3 > 0
    assert nobs3 > nobs1



#!/usr/bin/env python3

try:
    import random
    import typing

    import alifedata_phyloinformatics_convert as apc
    import opytional as opyt
    import pandas as pd

    from hstrat import _auxiliary_lib as hstrat_auxlib
    from hstrat import hstrat
except:
    print(
        """
        Missing dependencies!
        Run

        python3 -m pip install hstrat

        to install necessary dependencies.
        """
    )
    raise

# simple asexual evolutionary algorithm under drift conditions
# 100 synchronous generations with population size 10
def evolve_drift_synchronous(population: typing.List) -> typing.List:
    # synchronous generations
    for generation in range(100):
        parents = random.choices(population, k=len(population))
        population = [parent.CreateOffspring() for parent in parents]

    return population


# example genome
# bundles (arbitrary) functional genome content with hereditary stratigraph
# annotation
class SimpleAnnotatedGenome:

    content: float  # put functional genome content here
    annotation: hstrat.HereditaryStratigraphicColumn

    def __init__(
        self: "SimpleAnnotatedGenome",
        content_: typing.Optional[float] = None,
        annotation_: typing.Optional[
            hstrat.HereditaryStratigraphicColumn
        ] = None,
    ) -> None:
        # generate functional genome content randomly if not provided
        self.content = opyt.or_value(
            content_,
            random.uniform(0, 100),
        )
        # create new annotation if not provided
        self.annotation = opyt.or_value(
            annotation_,
            hstrat.HereditaryStratigraphicColumn(
                # specify which stratum retention policy to use...
                # this policy guarantees uncertainty no longer than 25% of
                # time elapsed since the true phylogenetic event
                hstrat.recency_proportional_resolution_algo.Policy(4),
                # how many bits for "fingerprints" (aka differentia)
                # generated & stored in each layer of column (uses 1 byte)
                stratum_differentia_bit_width=8,
            ),
        )

    def CreateOffspring(
        self: "SimpleAnnotatedGenome",
    ) -> "SimpleAnnotatedGenome":
        return SimpleAnnotatedGenome(
            # mutate functional genetic content
            self.content + random.uniform(0.0, 1.0) % 100.0,
            # tag offspring with an independent copy of the annotation, with a
            # new "fingerprint" layer appended
            self.annotation.CloneDescendant(),
        )


if __name__ == "__main__":

    # ensure reproducible results
    hstrat_auxlib.seed_random(2)

    # do the evolution
    print("evolving 100 generations...")
    population: typing.List[SimpleAnnotatedGenome]
    population = evolve_drift_synchronous(
        [SimpleAnnotatedGenome() for __ in range(10)]
    )

    # speciation event!
    # divide populations in two and evolve sub-populations separately
    print("speciation event! dividing population into two halves")
    print("evolving another 100 generations...")
    population = evolve_drift_synchronous(
        population[: len(population) // 2]
    ) + evolve_drift_synchronous(population[len(population) // 2 :])

    # evolution done
    print("evolution done!")
    # extract annotations from genomes
    extant_annotations = [genome.annotation for genome in population]

    # estimated_phylogeny is stored in alife data standards format
    # https://alife-data-standards.github.io/alife-data-standards/phylogeny.html
    estimated_phylogeny: pd.DataFrame
    estimated_phylogeny = hstrat.build_tree(
        extant_annotations,
        # the `build_tree` function tracks the current best-known general
        # purpose reconstruction algorithm
        # pin to the current version (e.g., "1.7.2") for long-term stability
        # or pin to hstrat.__version__ to track latest algorithm updates
        version_pin=hstrat.__version__,
    )
    # translate to dendropy (which provides lots of phylogenetics tools)
    # via alifedata phyloinformatics conversion tool
    dendropy_tree = apc.alife_dataframe_to_dendropy_tree(
        estimated_phylogeny,
        setup_edge_lengths=True,
    )

    # draw the reconstruction! (note detection of speciation event)
    print(dendropy_tree.as_ascii_plot(plot_metric="age"))



import numpy as np
import pandas as pd
import pytest

from hstrat._auxiliary_lib import get_nullable_mask, get_nullable_vals


@pytest.mark.parametrize("dtype_base", ["UInt{}", "Int{}"])
@pytest.mark.parametrize("dtype_bits", [8, 16, 32, 64])
def test_get_nullable_vals(dtype_base, dtype_bits):
    series = pd.Series([3, 1, 2], dtype=dtype_base.format(dtype_bits))
    assert np.array_equal(get_nullable_vals(series), series)
    get_nullable_vals(series)[0] = 42
    assert np.array_equal(get_nullable_vals(series), np.array([42, 1, 2]))
    get_nullable_mask(series)[2] = False
    assert get_nullable_vals(series)[2] == 2



import operator

import pytest

from hstrat import hstrat


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=4
        ),
    ],
)
@pytest.mark.parametrize(
    "num_deposits",
    [0, 1, 6, 8, 64, 65],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 2, 8, 20, 32, 64, 129],
)
def test_col_to_specimen(
    retention_policy,
    num_deposits,
    differentia_bit_width,
):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_bit_width,
    ).CloneNthDescendant(num_deposits)

    specimen = hstrat.col_to_specimen(column)
    assert all(
        map(
            lambda x: operator.eq(*x),
            zip(column.IterRetainedRanks(), specimen.GetRankIndex()),
        )
    )
    assert all(
        map(
            lambda x: operator.eq(*x),
            zip(
                column.IterRetainedDifferentia(), specimen.GetDifferentiaVals()
            ),
        )
    )



import numpy as np
import pytest

from hstrat.hstrat import fixed_resolution_algo
from hstrat.stratum_retention_strategy.stratum_retention_algorithms._impl import (
    CalcMrcaUncertaintyAbsUpperBoundPessimalRankBruteForce,
)


@pytest.mark.parametrize(
    "fixed_resolution",
    [
        1,
        2,
        3,
        7,
        42,
        100,
    ],
)
@pytest.mark.parametrize(
    "time_sequence",
    [
        np.random.default_rng(1).integers(
            10**2,
            size=10,
        ),
        np.random.default_rng(1).integers(
            10**3,
            size=10,
        ),
    ],
)
def test_policy_consistency(fixed_resolution, time_sequence):
    policy = fixed_resolution_algo.Policy(fixed_resolution)
    spec = policy.GetSpec()
    instance = (
        fixed_resolution_algo.CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank(
            spec,
        )
    )
    for num_strata_deposited_a in time_sequence:
        for num_strata_deposited_b in (
            num_strata_deposited_a // 3,
            num_strata_deposited_a // 2,
            num_strata_deposited_a,
            num_strata_deposited_a + 1,
            num_strata_deposited_a + 10,
            num_strata_deposited_a + 100,
        ):
            if 0 in (num_strata_deposited_a, num_strata_deposited_b):
                continue
            for actual_mrca_rank in np.random.default_rng(1).integers(
                min(num_strata_deposited_a, num_strata_deposited_b),
                size=3,
            ):
                policy_requirement = policy.CalcMrcaUncertaintyAbsUpperBound(
                    num_strata_deposited_a,
                    num_strata_deposited_b,
                    CalcMrcaUncertaintyAbsUpperBoundPessimalRankBruteForce()(
                        policy,
                        num_strata_deposited_a,
                        num_strata_deposited_b,
                    ),
                )
                for which in (
                    instance,
                    fixed_resolution_algo.CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank(
                        spec
                    ),
                ):
                    assert (
                        which(
                            policy,
                            num_strata_deposited_a,
                            num_strata_deposited_b,
                        )
                        == policy_requirement
                    )


@pytest.mark.parametrize(
    "fixed_resolution",
    [
        1,
        2,
        3,
        7,
        42,
        100,
    ],
)
def test_eq(fixed_resolution):
    policy = fixed_resolution_algo.Policy(fixed_resolution)
    spec = policy.GetSpec()
    instance = (
        fixed_resolution_algo.CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank(
            spec
        )
    )

    assert instance == instance
    assert (
        instance
        == fixed_resolution_algo.CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank(
            spec,
        )
    )
    assert instance is not None



import importlib
import logging
import os
import types

import opytional as opyt

from ._except_wrap_sentinel import except_wrap_sentinel
from ._is_in_unit_test import is_in_unit_test


def _import_cppimport(module_name: str, package: str) -> types.ModuleType:
    """Implementation detail for load_cppimportable_module"""
    import cppimport

    return cppimport.imp(f"{package}.{module_name}")


def _import_importlib(module_name: str, package: str) -> types.ModuleType:
    """Implementation detail for load_cppimportable_module"""
    return importlib.import_module(f".{module_name}", package=package)


def load_cppimportable_module(
    module_name: str, package: str
) -> types.ModuleType:
    r"""Imports module, prioritizing cppimport if in unit test or requested
    by environment variable HSTRAT_USE_CPPIMPORT; otherwise, pre-existing
    precompiled binaries are prioritized.

    Parameters
    ----------
    module_name : str
        The name of the C++ module to import from.
    package : str
        The package where the C++ module is contained.

        Assuming this function is called in the same package as that, it should
        be something like `re.sub(r"\.[a-za-z0-9_]+$", '', __name__)`.

    Returns
    -------
    types.ModuleType
        The imported module.
    """

    def do_import_importlib() -> types.ModuleType:
        return _import_importlib(module_name, package)

    def do_import_cppimport() -> types.ModuleType:
        return _import_cppimport(module_name, package)

    in_unit_test = is_in_unit_test()
    if "HSTRAT_USE_CPPIMPORT" in os.environ or in_unit_test:
        if in_unit_test:
            logging.info(
                "unit test session detected, preferring cppimport over precompiled binaries",
            )
        primary, fallback = (
            except_wrap_sentinel(
                do_import_cppimport,
                {
                    ImportError: f"cppimport failed, falling back to precompiled binaries for '{package}.{module_name}'",
                    ModuleNotFoundError: f"cppimport not installed, falling back to native binaries for '{package}.{module_name}'",
                },
                sentinel=None,
            ),
            do_import_importlib,
        )
    else:
        primary, fallback = (
            except_wrap_sentinel(
                do_import_importlib,
                {
                    ImportError: f"precompiled binaries for '{package}.{module_name}' not found, falling back to cppimport",
                },
            ),
            do_import_cppimport,
        )

    return opyt.or_else(primary(), fallback)



import json
import logging
import operator

import pytest

from hstrat import hstrat
from hstrat._auxiliary_lib import get_hstrat_version, log_once_in_a_row


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=4
        ),
    ],
)
@pytest.mark.parametrize(
    "num_deposits",
    [0, 1, 6, 8, 64, 65],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 2, 8, 20, 32, 64, 129],
)
def test_specimen_from_records1(
    retention_policy,
    num_deposits,
    differentia_bit_width,
):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_bit_width,
    ).CloneNthDescendant(num_deposits)

    specimen = hstrat.specimen_from_records(hstrat.col_to_records(column))
    assert all(
        map(
            lambda x: operator.eq(*x),
            zip(column.IterRetainedRanks(), specimen.GetRankIndex()),
        )
    )
    assert all(
        map(
            lambda x: operator.eq(*x),
            zip(
                column.IterRetainedDifferentia(), specimen.GetDifferentiaVals()
            ),
        )
    )


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=4
        ),
    ],
)
@pytest.mark.parametrize(
    "num_deposits",
    [0, 1, 6, 8, 64, 65],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 2, 8, 20, 32, 64],
)
def test_col_to_records_then_to_specimen_json(
    retention_policy,
    num_deposits,
    differentia_bit_width,
):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_bit_width,
    ).CloneNthDescendant(num_deposits)

    records = hstrat.col_to_records(column)
    json_str = json.dumps(records)
    reconstituted_records = json.loads(json_str)
    specimen = hstrat.specimen_from_records(reconstituted_records)
    assert all(
        map(
            lambda x: operator.eq(*x),
            zip(column.IterRetainedRanks(), specimen.GetRankIndex()),
        )
    )
    assert all(
        map(
            lambda x: operator.eq(*x),
            zip(
                column.IterRetainedDifferentia(), specimen.GetDifferentiaVals()
            ),
        )
    )


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=4
        ),
    ],
)
@pytest.mark.parametrize(
    "num_deposits",
    [0, 1, 6, 8, 64, 65],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 2, 8, 20, 32, 64],
)
def test_specimen_from_records2(
    retention_policy,
    num_deposits,
    differentia_bit_width,
    caplog,
):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_bit_width,
    ).CloneNthDescendant(num_deposits)

    # adapted from https://stackoverflow.com/a/48113200
    with caplog.at_level(logging.INFO):
        # clear caplog
        caplog.clear()

        # create a record with a phony version
        record = hstrat.col_to_records(column)
        record["hstrat_version"] = "0.0.0"

        # reconstruct phony record
        hstrat.specimen_from_records(record)

        # it should have logged one warning
        assert len(caplog.records) == 1
        # retrieve it
        logged = next(iter(caplog.records))

        assert logged.levelno == logging.INFO
        assert logged.name == "hstrat"
        assert logged.module == "_log_once_in_a_row"
        assert (
            logged.message
            == f"""specimen_from_records version mismatch, record is version {
                record["hstrat_version"]
            } and software is version {
                get_hstrat_version()
            }"""
        )

        # log something to keep the lru cache from supressing the above calls
        log_once_in_a_row("foo")



from copy import copy

import pandas as pd
import pytest

from hstrat._auxiliary_lib import CopyableSeriesItemsIter


@pytest.fixture
def sample_series():
    return pd.Series([10, 20, 30, 40], index=["a", "b", "c", "d"])


def test_copyable_series_items_iter_initial_position(sample_series):
    c = CopyableSeriesItemsIter(sample_series)
    assert c._position == 0


def test_copyable_series_items_iter_next_returns_tuple(sample_series):
    c = CopyableSeriesItemsIter(sample_series)
    assert isinstance(next(c), tuple)


def test_copyable_series_items_iter_iteration(sample_series):
    c = CopyableSeriesItemsIter(sample_series)
    items = list(c)
    assert len(items) == len(sample_series)
    assert items[0] == ("a", 10)
    assert items[-1] == ("d", 40)


def test_copyable_series_items_iter_empty_series():
    empty_series = pd.Series(dtype=int)
    c = CopyableSeriesItemsIter(empty_series)
    with pytest.raises(StopIteration):
        next(c)


def test_copyable_series_items_iter_copy1():
    s = pd.Series([1, 2, 3])
    c = CopyableSeriesItemsIter(s)
    c_copy = copy(c)
    assert isinstance(c_copy, CopyableSeriesItemsIter)
    assert list(c_copy) == list(c)


def test_copyable_series_items_iter_copy2():
    s = pd.Series([1, 2, 3])
    c = CopyableSeriesItemsIter(s)
    next(c)
    c_copy = copy(c)
    assert isinstance(c_copy, CopyableSeriesItemsIter)
    assert list(c_copy) == list(c)



#!/usr/bin/env python3
# note: temporarily disabled because failing on gh actions build as "cancelled"

import gzip
import io
import json
import os
import random
import tempfile

import yaml

from hstrat import hstrat

if __name__ == "__main__":

    policy = hstrat.recency_proportional_resolution_algo.Policy(3)

    common_ancestor = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
    )
    pop = [common_ancestor.CloneDescendant() for __ in range(10)]

    for __ in range(100):
        target = random.randrange(len(pop))
        pop[target] = random.choice(pop).CloneDescendant()

    print(
        "export population of hereditary stratigraphic columns "
        "to pandas dataframe"
    )
    print(hstrat.pop_to_dataframe(pop))
    print()

    # horizontal divider, adapted from https://stackoverflow.com/a/42762743
    os.system(r"""printf '%*s\n' "${COLUMNS:-$(tput cols)}" '' | tr ' ' -""")

    with tempfile.NamedTemporaryFile("w+") as tmp:
        print(f"serialize population as json to file {tmp.name}")
        json.dump(hstrat.pop_to_records(pop), tmp, indent=4)
        tmp.flush()
        os.system(f"set -x; head -c 500 {tmp.name} | head -n 10")
        print("...")
        print("show file size")
        os.system(f"set -x; du -b {tmp.name}")
        print()

        tmp.seek(0)
        deserialized = hstrat.pop_from_records(json.load(tmp))
        print(f"deserialized and original equal? {deserialized == pop}")
        print()

    # horizontal divider, adapted from https://stackoverflow.com/a/42762743
    os.system(r"""printf '%*s\n' "${COLUMNS:-$(tput cols)}" '' | tr ' ' -""")

    with tempfile.NamedTemporaryFile("w+") as tmp:
        print(f"serialize population as yaml to file {tmp.name}")
        yaml.dump(hstrat.pop_to_records(pop), tmp)
        tmp.flush()
        os.system(f"set -x; head -c 500 {tmp.name} | head -n 10")
        print("...")
        print("show file size")
        os.system(f"set -x; du -b {tmp.name}")
        print()

        tmp.seek(0)
        deserialized = hstrat.pop_from_records(yaml.safe_load(tmp))
        print(f"deserialized and original equal? {deserialized == pop}")
        print()

    # horizontal divider, adapted from https://stackoverflow.com/a/42762743
    os.system(r"""printf '%*s\n' "${COLUMNS:-$(tput cols)}" '' | tr ' ' -""")

    with tempfile.NamedTemporaryFile("w+b") as tmp:
        tmp_gz = gzip.GzipFile(mode="w+b", fileobj=tmp)
        print(f"serialize population as json to file {tmp.name} with gzip")
        json.dump(hstrat.pop_to_records(pop), io.TextIOWrapper(tmp_gz))
        tmp.flush()
        os.system(
            f"set -x; cat {tmp.name} | gunzip | head -c 500 | head -n 10"
        )
        print("...")
        print("show file size")
        os.system(f"set -x; du -b {tmp.name}")
        print()

        with gzip.open(tmp.name, "r") as tmp_gz_in:
            records = json.load(io.TextIOWrapper(tmp_gz_in))
            deserialized = hstrat.pop_from_records(records)
            print(f"deserialized and original equal? {deserialized == pop}")



from typing import Any, Callable, List, Tuple

import lazy_loader

from ._GetAttrLaunderShim import GetAttrLaunderShim


def lazy_attach_stub(
    module_name: str,
    module_path: str,
    *,
    should_launder: Callable = lambda _: True,
) -> Tuple[Callable[[str], Any], Callable[[], List[str]], List[str]]:
    """
    Attaches a lazy loading stub to a package, enabling deferred loading
    of submodules and attributes to improve load times. Used by importing
    necessary symbols into a type stub (a .pyi file) and evaluating said
    type stub to determine symbols to import. Called in the __init__.py
    file of a package, assuming there exists an __init__.pyi file too.

    Parameters
    ----------
    module_name : str
        The name of the package in which symbols are being imported. This
        should be `__name__` to properly set up the lazy loading stub.
    module_path : str
        The file path of the main module. This should be `__file__` to ensure
        correct module referencing.
    should_launder : callable, default `lambda x: True`
        A callable that takes a string attr and returns a boolean indicating
        whether the attr should be laundered. If True, __module__ on the
        attr will attempt to be set to `module_name`.

    Returns
    -------
    tuple
        A tuple containing the items needed to define `__getattr__`, `__dir__`,
        and `__all__` in the main module:

        - `__getattr__` : function
            Retrieves attributes lazily as needed.
        - `__dir__` : function
            Lists all available attributes, including those lazily loaded.
        - `__all__` : list of str
            A list of all attributes for the namespace to be included in *
            imports and accessbile from other modules.

    Notes
    -----
    This function relies on the lazy loading mechanism provided by
    `attach_stub` to minimize initial load times for large libraries by
    delaying the loading of specific modules and attributes until they
    are accessed.

    For additional usage information, refer to the Scientific Python
    project documentation:
    https://scientific-python.org/specs/spec-0001/

    See Also
    --------
    ._lazy_attach : Provides an alternative lazy loading method,
        where submodules and attributes are declared directly in the
        __init__.py file. This was used where `__all__` imports are forwarded from
        subpackages, as lazy_loader does not support that natively.
    ._launder_impl_modules : Implements laundering for module names
    in attributes.
    """
    getattr__, dir__, all__ = lazy_loader.attach_stub(module_name, module_path)
    wrapped_getattr = GetAttrLaunderShim(
        getattr__, should_launder, module_name
    )
    return wrapped_getattr, dir__, all__



import functools
import typing
import warnings


def except_wrap_sentinel(
    func: typing.Callable,
    errors: dict[typing.Type[Exception], typing.Optional[str]],
    *,
    sentinel: typing.Any = None,
) -> typing.Callable:
    """Decorator to catch specified exceptions and return a sentinel value.

    Parameters
    ----------
    func : callable
        The function to wrap.
    errors : dict of Exception to str or None
        A mapping of exception types to optional warning messages. If the
        function raises one of these exceptions, that exception will be caught.
        If a message is provided, a warning will be issued. If no message is
        provided (None), no warning will be issued.
    sentinel : object, optional
        The value to return if any of the specified exceptions is raised, by
        default None.

    Returns
    -------
    callable
        A wrapped version of `func`. When called, if `func` raises one of the
        exceptions listed in `errors`, `sentinel` will be returned instead, and
        a warning may be issued if specified.
    """

    @functools.wraps(func)
    def inner(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            for err_type, message in errors.items():
                if isinstance(e, err_type):
                    if message is not None:
                        warnings.warn(message)
                    return sentinel
            raise e

    return inner



from collections import Counter

import pandas as pd

from ._alifestd_parse_ancestor_ids import alifestd_parse_ancestor_ids


def alifestd_topological_sort(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Sort rows so all organisms follow members of their `ancestor_list`.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df.set_index("id", drop=False, inplace=True)

    unsorted_ids = set(phylogeny_df["id"])
    internal_ids_refcounts = Counter(
        id
        for ancestor_list_str in phylogeny_df["ancestor_list"]
        for id in alifestd_parse_ancestor_ids(ancestor_list_str)
    )
    internal_ids_set = set(internal_ids_refcounts.keys())

    reverse_sorted_ids = []
    while unsorted_ids:
        leaf_ids = unsorted_ids - internal_ids_set
        reverse_sorted_ids.extend(leaf_ids)
        unsorted_ids -= leaf_ids

        leaf_parent_counts = Counter(
            id
            for ancestor_list_str in phylogeny_df.loc[[*leaf_ids]][
                "ancestor_list"
            ]
            for id in alifestd_parse_ancestor_ids(ancestor_list_str)
        )
        internal_ids_refcounts.subtract(leaf_parent_counts)

        for leaf_parent_id in leaf_parent_counts.keys():
            refcount = internal_ids_refcounts[leaf_parent_id]
            assert refcount >= 0
            if refcount == 0:
                internal_ids_set.remove(leaf_parent_id)
                del internal_ids_refcounts[leaf_parent_id]

    return phylogeny_df.loc[reverse_sorted_ids[::-1], :].reset_index(drop=True)



"""Functors to specify property policies should be parameterized for."""

from ..._auxiliary_lib import lazy_attach_stub

__getattr__, __dir__, __all__ = lazy_attach_stub(__name__, __file__)
del lazy_attach_stub



import typing

from iterpop import iterpop as ip

from ..genome_instrumentation import (
    HereditaryStratigraphicColumn,
    HereditaryStratum,
)
from ._calc_rank_of_last_retained_commonality_between import (
    calc_rank_of_last_retained_commonality_between,
)


def get_last_common_stratum_between(
    first: HereditaryStratigraphicColumn,
    second: HereditaryStratigraphicColumn,
    confidence_level: float = 0.95,
) -> typing.Optional[HereditaryStratum]:
    """Get the most recent stratum in common between first and second, if any.

    Common strata share identical rank and differentia. Returns None if no
    common strata exist between the two columns. Allows probability equal
    to 1 - confidence_level that the last true common stratum is before the
    stratum returned (i.e., strata were erroneously detected as common due
    to spurious differentia collisions).

    See Also
    --------
    calc_rank_of_last_retained_commonality_between :
        Selects the stratum returned. See the corresponding docstring for
        explanation of parameters.
    """
    rank = calc_rank_of_last_retained_commonality_between(
        first,
        second,
        confidence_level=confidence_level,
    )
    if rank is not None:
        index = ip.popsingleton(
            index
            for index in range(first.GetNumStrataRetained())
            if rank == first.GetRankAtColumnIndex(index)
        )
        return first.GetStratumAtColumnIndex(index)
    else:
        return None



"""Implementation helpers."""

from ._estimate_rank_of_mrca_maximum_likelihood import (
    estimate_rank_of_mrca_maximum_likelihood,
)
from ._estimate_rank_of_mrca_naive import estimate_rank_of_mrca_naive
from ._estimate_rank_of_mrca_unbiased import estimate_rank_of_mrca_unbiased
from ._extract_common_retained_ranks_through_first_retained_disparity import (
    extract_common_retained_ranks_through_first_retained_disparity,
)

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "estimate_rank_of_mrca_maximum_likelihood",
    "estimate_rank_of_mrca_naive",
    "estimate_rank_of_mrca_unbiased",
    "extract_common_retained_ranks_through_first_retained_disparity",
]



def join_paragraphs_from_one_sentence_per_line(text: str) -> str:
    """Group lines of text into paragraphs, assuming one sentence per line.

    Strips newline after each paragraph.
    """
    return text.replace(".\n", ". ")



import typing

from ._detail import policy_param_focalizer_t


class NumStrataRetainedUpperBoundEvaluator:
    """Enacts space complexity bound parameterization requirement."""

    _at_num_strata_deposited: int
    _policy_param_focalizer: policy_param_focalizer_t

    def __init__(
        self: "NumStrataRetainedUpperBoundEvaluator",
        at_num_strata_deposited: int,
        policy_param_focalizer: policy_param_focalizer_t = (
            lambda policy_t: lambda i: policy_t(i)
        ),
    ) -> None:
        """Initialize functor to evaluate upper bound on num strata retained.

        Parameters
        ----------
        at_num_strata_deposited : int
            At what generation should we evaluate policy?
        policy_param_focalizer : callable
            Callable to create shim that constructs policy instance from
            parameter value. Default passes parameter value as sole argument to
            policy constructor.
        """
        self._at_num_strata_deposited = at_num_strata_deposited
        self._policy_param_focalizer = policy_param_focalizer

    def __call__(
        self: "NumStrataRetainedUpperBoundEvaluator",
        policy_t: typing.Type,
        parameter_value: int,
    ) -> int:
        """Get upper bound on num strata retained for specific parameter."""
        policy_factory = self._policy_param_focalizer(policy_t)
        policy = policy_factory(parameter_value)
        return policy.CalcNumStrataRetainedUpperBound(
            self._at_num_strata_deposited,
        )

    def __repr__(self: "NumStrataRetainedUpperBoundEvaluator") -> str:
        return f"""{
            NumStrataRetainedUpperBoundEvaluator.__qualname__
        }(at_num_strata_deposited={
            self._at_num_strata_deposited
        !r}, policy_param_focalizer={
            self._policy_param_focalizer
        !r})"""

    def __str__(self: "NumStrataRetainedUpperBoundEvaluator") -> str:
        title = "Upper Bound Num Strata Retained Evaluator"
        return f"""{
            title
        } (at num strata deposited: {
            self._at_num_strata_deposited
        }, focalizer: {
            self._policy_param_focalizer
        })"""



import pandas as pd

from ._alifestd_find_chronological_inconsistency import (
    alifestd_find_chronological_inconsistency,
)
from ._alifestd_is_topologically_sorted import alifestd_is_topologically_sorted
from ._alifestd_parse_ancestor_ids import alifestd_parse_ancestor_ids
from ._alifestd_topological_sort import alifestd_topological_sort


def alifestd_coerce_chronological_consistency(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """For any taxa with origin time preceding its parent's, set origin time
    to parent's origin time.

    If an inconsistency is detected, the corrected phylogeny will be returned
    sorted in topological order.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    if (
        not alifestd_is_topologically_sorted(phylogeny_df)
        and alifestd_find_chronological_inconsistency(phylogeny_df) is not None
    ):
        phylogeny_df = alifestd_topological_sort(phylogeny_df, mutate=True)

    phylogeny_df = phylogeny_df.set_index("id", drop=False)

    # adapted from https://stackoverflow.com/a/31569794
    origin_time_loc = phylogeny_df.columns.get_loc("origin_time")
    for pos, (_idx, row) in enumerate(phylogeny_df.iterrows()):
        for ancestor_id in alifestd_parse_ancestor_ids(row["ancestor_list"]):
            ancestor_loc = phylogeny_df.index.get_loc(ancestor_id)
            ancestor_time = phylogeny_df.iat[ancestor_loc, origin_time_loc]
            offspring_time = phylogeny_df.iat[pos, origin_time_loc]
            phylogeny_df.iat[pos, origin_time_loc] = max(
                offspring_time, ancestor_time
            )

    return phylogeny_df



import typing

import interval_search as inch

from ....._auxiliary_lib import div_range, memoize_generator
from ._calc_rank_backstop import calc_rank_backstop
from ._calc_rank_sep import calc_rank_sep


@memoize_generator()
def iter_priority_ranks(
    degree: int,
    interspersal: int,
    pow_: int,
    num_strata_deposited: int,
) -> typing.Iterator[int]:
    """Iterate over retained ranks for the `pow`th target recency.

    Ranks are yielded from last-to-be-deleted to first-to-be-deleted. The
    very-first-to-be-deleted (i.e., lowest priority) rank is yielded last. Will
    yield all ranks less than `num_strata_deposited` before exhaustion.

    Notes
    -----
    May yield duplicates (i.e., yield rank x then later again yield rank x).
    This is acceptable in the internal use case because rank priority only
    depends on its soonest position in the stream of yielded values.
    Subsequent yieldings have no effect because the generator is sampled until
    exhausted or an unseen rank is yielded.
    """
    # aliases for terseness
    d = degree
    i = interspersal

    min_retained_rank = calc_rank_backstop(d, i, pow_, num_strata_deposited)
    retained_ranks_sep = calc_rank_sep(d, i, pow_, num_strata_deposited)

    if pow_ == 0:
        # optimization
        yield from reversed(range(num_strata_deposited))
        return
    elif pow_ == degree:
        # the highest-degree pow always tracks strata 0
        # and the biggest relevant sep for strata 0 is infinite.
        # i.e., the backstop for highest-debree pow is always strata 0
        # So, we need a special case for this pow.
        # However, because strata 0 is retained indefinitely, we actually
        # only need to worry about the next retained strata,
        # which is at retained_ranks_sep.
        # Thus, we use calc_rank_sep instead of calc_rank_backstop.
        # TODO can this doubling search be done in constant time?
        biggest_relevant_rank = inch.doubling_search(
            lambda x: calc_rank_sep(d, i, pow_, x + 1) >= num_strata_deposited,
            num_strata_deposited,
        )
        biggest_relevant_sep = calc_rank_sep(
            degree,
            interspersal,
            pow_,
            biggest_relevant_rank,
        )

    else:
        # TODO can this doubling search be done in constant time?
        biggest_relevant_rank = inch.doubling_search(
            lambda x: calc_rank_backstop(d, i, pow_, x + 1)
            >= num_strata_deposited,
            num_strata_deposited,
        )
        biggest_relevant_sep = calc_rank_sep(
            degree,
            interspersal,
            pow_,
            biggest_relevant_rank,
        )

    # in practice, just "cur_sep == retained_ranks * 2" appears required
    # i.e., tests pass with
    #
    # for cur_sep in retained_ranks_sep * 2,:
    #
    # TODO can this be proven?
    for cur_sep in div_range(
        biggest_relevant_sep,  # start
        retained_ranks_sep,  # stop, non-inclusive
        2,  # iteration action: divide by 2
    ):
        # TODO can this doubling search be done in constant time?
        cur_sep_rank = inch.doubling_search(
            lambda x: calc_rank_sep(d, i, pow_, x) >= cur_sep,
            # cur_sep is always guaranteed at least leq its threshold
            # num_strata_deposited, so we can safely begin searching for its
            # threshold at cur_sep
            # (look at calc_rank_sep to convince yourself)
            # this optimization reduces necessary search steps and prevents
            # exceedance of maximum recursion depth in some tests
            max(num_strata_deposited, cur_sep),
        )
        cur_sep_rank_backstop = calc_rank_backstop(
            degree,
            interspersal,
            pow_,
            cur_sep_rank,
        )

        yield from reversed(
            range(
                cur_sep_rank_backstop,  # start
                # +1 to be inclusive of cur_sep_rank
                min(cur_sep_rank + 1, num_strata_deposited),  # stop
                cur_sep,  # sep
            )
        )

    # TODO somehow exclude duplicates with above for better efficiency?
    yield from reversed(
        range(
            min_retained_rank,  # start
            num_strata_deposited,  # stop, non-inclusive
            retained_ranks_sep,  # sep
        )
    )

    # recurse
    if retained_ranks_sep == 1:
        # base case
        yield from reversed(
            range(
                0,
                min_retained_rank,
            )
        )
        return

    prev_sep_rank = inch.binary_search(
        lambda x: calc_rank_sep(d, i, pow_, x + 1) >= retained_ranks_sep,
        0,
        num_strata_deposited - 1,
    )
    yield from range(
        min_retained_rank,  # start
        prev_sep_rank,  # stop, not inclusive
        -retained_ranks_sep,  # sep
    )
    assert prev_sep_rank < num_strata_deposited
    yield from iter_priority_ranks(
        degree,
        interspersal,
        pow_,
        # + 1 due to apparent off-by-one error w/ just prev_sep_rank
        # where rank 5984 isn't properly retained
        # with degree 9, interspersal 2 @ generation 6405
        # on get_retained_ranks test case
        min(prev_sep_rank + 1, num_strata_deposited - 1),
    )



import typing

from .._auxiliary_lib import HereditaryStratigraphicArtifact
from ._impl import dispatch_impl


def calc_rank_of_last_retained_commonality_between(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
    confidence_level: float = 0.95,
) -> typing.Optional[int]:
    """Determine lower bound on generation of MRCA at confidence level.

    How many depositions elapsed along the columns' lines of
    descent before the last matching strata at the same rank between
    first and second?

    Parameters
    ----------
    confidence_level : float, optional
        With what probability should the true rank of the last commonality
        between first and second fall at or after the returned rank? Default
        0.95.

    Returns
    -------
    int, optional
        The number of depositions elapsed or None if no common ancestor is
        shared between the columns.

    Notes
    -----
    The true rank of the last commonality with second is guaranteed to never
    be after the returned rank when confidence_level < 0.5.
    """
    assert 0.0 <= confidence_level <= 1.0

    return dispatch_impl(
        first, second
    ).calc_rank_of_last_retained_commonality_between(
        first, second, confidence_level
    )



from collections import deque
import itertools as it
import typing


def flag_last(
    iterable: typing.Iterable,
) -> typing.Iterator[typing.Tuple[bool, typing.Any]]:
    """Yields tuples of the form (is_last, value) for each value in the
    iterable, where is_last is True if the value is the last in the iterable
    and False otherwise.

    Parameters
    ----------
    iterable : iterable
        The iterable to be processed.

    Yields
    ------
    flagged_value : tuple
        Tuples of the form (is_last, value) for each value in the iterable.

    Examples
    --------
    >>> [*flag_last([1, 2, 3, 4, 5])]
    [(False, 1), (False, 2), (False, 3), (False, 4), (True, 5)]

    >>> [*flag_last([1])]
    [(True, 1)]

    >>> [*flag_last([])]
    []

    """
    iterator = iter(iterable)
    lookahead = deque(it.islice(iterator, 2), maxlen=2)

    while lookahead:
        yield (len(lookahead) == 1, lookahead.popleft())
        for val in iterator:
            lookahead.append(val)
            break



from ._ballpark_patristic_distance_between import (
    ballpark_patristic_distance_between,
)
from ._ballpark_rank_of_mrca_between import ballpark_rank_of_mrca_between
from ._ballpark_ranks_since_mrca_with import ballpark_ranks_since_mrca_with
from ._calc_patristic_distance_bounds_between import (
    calc_patristic_distance_bounds_between,
)
from ._calc_rank_of_earliest_detectable_mrca_between import (
    calc_rank_of_earliest_detectable_mrca_between,
)
from ._calc_rank_of_mrca_bounds_between import calc_rank_of_mrca_bounds_between
from ._calc_rank_of_mrca_bounds_provided_confidence_level import (
    calc_rank_of_mrca_bounds_provided_confidence_level,
)
from ._calc_rank_of_mrca_uncertainty_between import (
    calc_rank_of_mrca_uncertainty_between,
)
from ._calc_ranks_since_earliest_detectable_mrca_with import (
    calc_ranks_since_earliest_detectable_mrca_with,
)
from ._calc_ranks_since_mrca_bounds_provided_confidence_level import (
    calc_ranks_since_mrca_bounds_provided_confidence_level,
)
from ._calc_ranks_since_mrca_bounds_with import (
    calc_ranks_since_mrca_bounds_with,
)
from ._calc_ranks_since_mrca_uncertainty_with import (
    calc_ranks_since_mrca_uncertainty_with,
)
from ._does_definitively_have_no_common_ancestor import (
    does_definitively_have_no_common_ancestor,
)
from ._does_have_any_common_ancestor import does_have_any_common_ancestor
from ._estimate_patristic_distance_between import (
    estimate_patristic_distance_between,
)
from ._estimate_rank_of_mrca_between import estimate_rank_of_mrca_between
from ._estimate_ranks_since_mrca_with import estimate_ranks_since_mrca_with

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "ballpark_patristic_distance_between",
    "ballpark_rank_of_mrca_between",
    "ballpark_ranks_since_mrca_with",
    "calc_patristic_distance_bounds_between",
    "calc_rank_of_earliest_detectable_mrca_between",
    "calc_rank_of_mrca_bounds_between",
    "calc_rank_of_mrca_bounds_provided_confidence_level",
    "calc_rank_of_mrca_uncertainty_between",
    "calc_ranks_since_earliest_detectable_mrca_with",
    "calc_ranks_since_mrca_bounds_with",
    "calc_ranks_since_mrca_bounds_provided_confidence_level",
    "calc_ranks_since_mrca_uncertainty_with",
    "does_definitively_have_no_common_ancestor",
    "does_have_any_common_ancestor",
    "estimate_patristic_distance_between",
    "estimate_rank_of_mrca_between",
    "estimate_ranks_since_mrca_with",
]



import typing

from .._auxiliary_lib import HereditaryStratigraphicArtifact
from ._calc_min_implausible_spurious_consecutive_differentia_collisions_between import (
    calc_min_implausible_spurious_consecutive_differentia_collisions_between,
)
from ._calc_rank_of_last_retained_commonality_between import (
    calc_rank_of_last_retained_commonality_between,
)


def calc_definitive_max_rank_of_last_retained_commonality_between(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
) -> typing.Optional[int]:
    """Determine latest possible generation of MRCA.

    At most, how many depositions elapsed along the columns' lines of
    descent before the last matching strata at the same rank between
    first and second?

    Returns
    -------
    int, optional
        The number of depositions elapsed or None if no common ancestor is
        shared between the columns.
    """
    confidence_level = 0.49
    assert (
        first.GetStratumDifferentiaBitWidth()
        == second.GetStratumDifferentiaBitWidth()
    )
    assert (
        calc_min_implausible_spurious_consecutive_differentia_collisions_between(
            first,
            second,
            significance_level=1.0 - confidence_level,
        )
        == 1
    )
    return calc_rank_of_last_retained_commonality_between(
        first,
        second,
        confidence_level=confidence_level,
    )



from .._auxiliary_lib import HereditaryStratigraphicArtifact


def calc_probability_differentia_collision_between(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
) -> float:
    """How likely are differentia collisions?

    Calculates the probability of two randomly-differentiated differentia
    being identical by coincidence.
    """
    assert (
        first.GetStratumDifferentiaBitWidth()
        == second.GetStratumDifferentiaBitWidth()
    )
    return 1.0 / 2 ** first.GetStratumDifferentiaBitWidth()



"""Top-level package for hstrat."""

__author__ = "Matthew Andres Moreno"
__email__ = "m.more500@gmail.com"
__version__ = "1.15.4"


from ._auxiliary_lib import lazy_attach_stub

__getattr__, __dir__, __all__ = lazy_attach_stub(
    __name__,
    __file__,
    should_launder=[].__contains__,
)
del lazy_attach_stub



#!/usr/bin/env python3

import argparse
import functools
import gc
import os
import random
import types
import typing
import uuid

import downstream
from downstream import dstream
import numpy as np
import opytional as opyt
import pandas as pd
from tqdm import tqdm

try:
    from phylotrackpy import systematics
except (ImportError, ModuleNotFoundError) as e:
    print("phylotrackpy required for dstream surf evolution example")
    print("python3 -m pip install phylotrackpy")
    raise e


def make_uuid4_fast() -> str:
    """Fast UUID4 generator, using lower-quality randomness."""
    return str(uuid.UUID(int=random.getrandbits(128), version=4))


def evolve_drift(population: typing.List) -> typing.List:
    """Simple asexual evolutionary algorithm under drift conditions."""
    selector = random.Random(1)  # ensure consistent true phylogeny

    # synchronous generations
    for generation in tqdm(range(500)):
        population = [
            parent.CreateOffspring()
            for parent in selector.choices(population, k=len(population))
        ]

    # asyncrhonous generations
    nsplit = len(population) // 2
    for generation in tqdm(range(500)):
        population[:nsplit] = [
            parent.CreateOffspring()
            for parent in selector.choices(population[:nsplit], k=nsplit)
        ]
        selector.shuffle(population)

    return population


def make_Organism(
    dstream_algo: types.ModuleType,
    differentia_bitwidth: int,
    surface_size: int,
    syst: systematics.Systematics,
) -> typing.Type:
    """Factory function to configure Organism class."""

    surf_dtype = {
        1: np.uint8,
        8: np.uint8,
        64: np.uint64,
    }[differentia_bitwidth]
    empty_surface = np.empty(surface_size, dtype=surf_dtype)
    empty_surface.setflags(write=False)

    surface_bitwidth = differentia_bitwidth * surface_size

    assign_site = functools.lru_cache(dstream_algo.assign_storage_site)
    algo_name = dstream_algo.__name__.split(".")[-1]

    class Organism:
        """Simple organism class, with instrumentation for both phylotrackpy
        and hstrat surface phylogeny tracking."""

        __slots__ = [
            "trait",
            "uid",
            "taxon",
            "generation_count",
            "hstrat_surface",
        ]

        # primary simulation business --- arbitrary data in this simple example
        trait: float

        # instrumentation: perfect tracking with phylotrackpy
        uid: str
        taxon: systematics.Taxon

        # instrumentation: approximate tracking with hstrat surface
        generation_count: int
        hstrat_surface: np.ndarray

        def __init__(
            self: "Organism",
            parent_taxon: typing.Optional[systematics.Taxon] = None,
            parent_hstrat_surface: np.ndarray = empty_surface,
            generation_count: int = 0,
            trait: float = 0.0,
        ) -> None:
            """Initialize organism, by default as root organism."""
            # handle primary simulation business...
            self.trait = (
                trait  # arbitrary example data, not used in simulation
            )

            # handle phylotrackpy instrumentation...
            self.uid = make_uuid4_fast()
            self.taxon = syst.add_org(self, parent_taxon)

            # handle hstrat surface instrumentation...
            self.generation_count = generation_count
            self.hstrat_surface = parent_hstrat_surface.copy()
            # ... deposit stratum...
            differentia_value = random.randrange(2**differentia_bitwidth)
            self.DepositStratum(differentia_value)

        def __del__(self: "Organism") -> None:
            """Remove organism from phylotrackpy systematics."""
            syst.remove_org(self.taxon)

        def CreateOffspring(self: "Organism") -> "Organism":
            """Create an offspring organism, with mutation and
            generation-updated instrumentation."""
            return Organism(
                parent_taxon=self.taxon,
                parent_hstrat_surface=self.hstrat_surface.copy(),
                generation_count=self.generation_count + 1,
                trait=self.trait + np.random.uniform(-1, 1),
            )

        def DepositStratum(self: "Organism", differentia_value: int) -> None:
            assert dstream_algo.has_ingest_capacity(
                surface_size, self.generation_count + 1
            )

            dstream_site = assign_site(surface_size, self.generation_count)
            if dstream_site is not None:  # handle skip/discard case
                self.hstrat_surface[dstream_site] = differentia_value

        def ToHex(self: "Organism") -> str:
            """Serialize the organism to a hex string, for genome output."""
            T_arr = np.asarray(self.generation_count + 1, dtype=np.uint32)
            T_bytes = T_arr.astype(">u4").tobytes()  # big-endian u32
            T_hex = T_bytes.hex()

            differentia_bytewidth = differentia_bitwidth // 8
            pack_op = [
                lambda x: x.astype(f">u{differentia_bytewidth}"),  # big-endian
                np.packbits,  # default big bitorder
            ][differentia_bitwidth == 1]
            surface_bits = pack_op(self.hstrat_surface)
            surface_bytes = surface_bits.tobytes()
            surface_hex = surface_bytes.hex()

            trait_arr = np.asarray(self.trait, dtype=np.float32)
            trait_bytes = trait_arr.tobytes()
            trait_hex = trait_bytes.hex()

            return T_hex + surface_hex + trait_hex

        def ToRecord(self: "Organism") -> dict:
            """Serialize the organism to a dictionary."""
            return {
                "downstream_version": downstream.__version__,
                "data_hex": self.ToHex(),
                "taxon_label": self.uid,
                "generation_count": self.generation_count,
                "dstream_algo": f"dstream.{algo_name}",
                "dstream_storage_bitoffset": 32,
                "dstream_storage_bitwidth": surface_bitwidth,
                "dstream_T_bitoffset": 0,
                "dstream_T_bitwidth": 32,
                "dstream_S": surface_size,
                "trait": self.trait,
            }

    return Organism


def make_validation_record(
    Organism: typing.Type,
    n_gen: int,
    differentia_override: typing.Callable,
    validator_exploded: str,
    validator_unpacked: str,
) -> dict:
    """Generate ephemeral validation data for quality assurance."""
    organism = None
    for T in range(n_gen):
        organism = opyt.apply_if_or_else(
            organism,
            Organism.CreateOffspring,
            Organism,
        )
        organism.DepositStratum(differentia_override(T))

    return {
        **organism.ToRecord(),
        "downstream_exclude_exploded": True,
        "downstream_validate_exploded": validator_exploded,
        "downstream_validate_unpacked": validator_unpacked,
    }


def _parse_args() -> argparse.Namespace:
    """Helper function to parse command line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument("--differentia-bitwidth", type=int, default=1)
    parser.add_argument("--surface-size", type=int, default=64)
    parser.add_argument(
        "--genome-df-path",
        type=str,
        default="/tmp/genome-evolve_surf_dstream.pqt",
    )
    parser.add_argument(
        "--phylo-df-path",
        type=str,
        default="/tmp/phylo-evolve_surf_dstream.csv",
    )

    args = parser.parse_args()

    if args.differentia_bitwidth not in (1, 8, 64):
        raise NotImplementedError()

    if args.surface_size < 8:
        raise NotImplementedError()

    if not args.phylo_df_path.endswith(".csv"):
        raise ValueError("phylo_df_path must end with '.csv'")

    return args


def _get_df_save_handler(path: str) -> typing.Callable:
    """Helper function to get the appropriate pandas dataframe save handler."""
    df_save_handlers = {
        ".csv": pd.DataFrame.to_csv,
        ".json": pd.DataFrame.to_json,
        ".pqt": pd.DataFrame.to_parquet,
        ".parquet": pd.DataFrame.to_parquet,
    }
    ext = os.path.splitext(path)[-1]
    try:
        return df_save_handlers[ext]
    except KeyError:
        raise ValueError(f"Unsupported dataframe path extension: {ext}")


if __name__ == "__main__":
    np.random.seed(1)  # ensure reproducibility
    random.seed(1)

    args = _parse_args()

    # configure organism class
    syst = systematics.Systematics(lambda x: x.uid)  # each org is own taxon
    syst.add_snapshot_fun(systematics.Taxon.get_info, "taxon_label")
    Organism = make_Organism(
        dstream_algo=dstream.steady_algo,
        differentia_bitwidth=args.differentia_bitwidth,
        surface_size=args.surface_size,
        syst=syst,
    )

    # do simulation
    common_ancestor = Organism()
    init_population = [common_ancestor.CreateOffspring() for _ in range(100)]
    end_population = evolve_drift(init_population)

    # mark non-tip taxa extinct
    del common_ancestor
    del init_population
    gc.collect()

    # set up validators to test during downstream processing
    S = args.surface_size
    checkerboard_validator = (
        "pl.col('dstream_value') == pl.col('dstream_Tbar') % 2"
    )
    block_validator = (
        f"pl.col('dstream_value') == pl.col('dstream_Tbar') // {S // 2 + 1}"
    )

    # write out the final population, including hstrat surface data
    genome_records = [
        *map(Organism.ToRecord, end_population),  # experiment data
        make_validation_record(  # ephemeral validation data
            Organism=Organism,
            n_gen=S,
            differentia_override=lambda T: T % 2,
            validator_exploded=checkerboard_validator,
            validator_unpacked=f"pl.col('dstream_T') == {S}",
        ),
        make_validation_record(  # ephemeral validation data
            Organism=Organism,
            n_gen=S + 1,
            differentia_override=lambda T: T % 2,
            validator_exploded=checkerboard_validator,
            validator_unpacked=f"pl.col('dstream_T') == {S + 1}",
        ),
        make_validation_record(  # ephemeral validation data
            Organism=Organism,
            n_gen=S,
            differentia_override=lambda T: T // (S // 2 + 1),
            validator_exploded=block_validator,
            validator_unpacked=f"pl.col('dstream_T') == {S}",
        ),
        make_validation_record(  # ephemeral validation data
            Organism=Organism,
            n_gen=S + 1,
            differentia_override=lambda T: T // (S // 2 + 1),
            validator_exploded=block_validator,
            validator_unpacked=f"pl.col('dstream_T') == {S + 1}",
        ),
    ]
    genome_df = pd.DataFrame(genome_records)
    _get_df_save_handler(args.genome_df_path)(genome_df, args.genome_df_path)

    # write out ground truth phylogeny
    syst.snapshot(args.phylo_df_path)



import anytree

from hstrat import hstrat
import hstrat.phylogenetic_inference.tree._impl as impl
import hstrat.phylogenetic_inference.tree.trie_postprocess._detail as detail


def test_base_class():
    assert issubclass(
        hstrat.AssignOriginTimeNodeRankTriePostprocessor,
        detail.TriePostprocessorBase,
    )


def test_assign_trie_origin_times_node_rank_single_leaf():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    leaf = impl.TrieLeafNode(parent=inner, taxon_label="A")
    root = hstrat.AssignOriginTimeNodeRankTriePostprocessor()(
        root, p_differentia_collision=0.5, mutate=True
    )
    assert leaf.origin_time == 0
    assert inner.origin_time == 0
    assert root.origin_time == 0


def test_assign_trie_origin_times_node_rank_two_leaves():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    leaf1 = impl.TrieLeafNode(parent=inner, taxon_label="A")
    leaf2 = impl.TrieLeafNode(parent=inner, taxon_label="B")
    root = hstrat.AssignOriginTimeNodeRankTriePostprocessor()(
        root, p_differentia_collision=0.5, mutate=True
    )
    assert root.origin_time == 0
    assert inner.origin_time == 0
    assert leaf1.origin_time == 0
    assert leaf2.origin_time == 0


def test_assign_trie_origin_times_node_rank1():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner1 = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    inner2 = impl.TrieInnerNode(rank=1, differentia=10, parent=inner1)
    leaf = impl.TrieLeafNode(parent=inner2, taxon_label="A")
    root = hstrat.AssignOriginTimeNodeRankTriePostprocessor()(
        root, p_differentia_collision=0.5, mutate=True
    )
    assert root.origin_time == 0
    assert inner1.origin_time == 0
    assert inner2.origin_time == 1
    assert leaf.origin_time == 1


def test_assign_trie_origin_times_node_rank2():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner1 = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    inner2 = impl.TrieInnerNode(rank=2, differentia=10, parent=inner1)
    leaf = impl.TrieLeafNode(parent=inner2, taxon_label="A")
    root = hstrat.AssignOriginTimeNodeRankTriePostprocessor()(
        root, p_differentia_collision=0.5, mutate=True
    )
    assert root.origin_time == 0
    assert inner1.origin_time == 0
    assert inner2.origin_time == 2
    assert leaf.origin_time == 2


def test_assign_trie_origin_times_node_rank3():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner1 = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    inner2 = impl.TrieInnerNode(rank=3, differentia=10, parent=inner1)
    leaf = impl.TrieLeafNode(parent=inner2, taxon_label="A")
    root = hstrat.AssignOriginTimeNodeRankTriePostprocessor()(
        root, p_differentia_collision=0.5, mutate=True
    )
    assert root.origin_time == 0
    assert inner1.origin_time == 0
    assert inner2.origin_time == 3
    assert leaf.origin_time == 3


def test_assign_trie_origin_times_node_rank_complex():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner1 = impl.TrieInnerNode(rank=0, differentia=0, parent=root)

    inner2 = impl.TrieInnerNode(rank=4, differentia=10, parent=inner1)
    inner3 = impl.TrieInnerNode(rank=7, differentia=7, parent=inner2)

    inner2a = impl.TrieInnerNode(rank=1, differentia=1, parent=inner1)

    leaf3_a = impl.TrieLeafNode(parent=inner3, taxon_label="leaf3a_a")

    leaf2a_a = impl.TrieLeafNode(parent=inner2a, taxon_label="leaf2a_a")
    leaf2a_b = impl.TrieLeafNode(parent=inner2a, taxon_label="leaf2a_b")

    root = hstrat.AssignOriginTimeNodeRankTriePostprocessor()(
        root, p_differentia_collision=0.5, mutate=True
    )

    assert root.origin_time == 0
    assert inner1.origin_time == 0
    assert inner2.origin_time == 4
    assert inner3.origin_time == 7
    assert inner2a.origin_time == 1
    assert leaf2a_a.origin_time == 1
    assert leaf2a_b.origin_time == 1
    assert leaf3_a.origin_time == 7


def test_assign_trie_origin_times_node_rank_assigned_property():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    leaf = impl.TrieLeafNode(parent=inner, taxon_label="A")
    root = hstrat.AssignOriginTimeNodeRankTriePostprocessor(
        assigned_property="blueberry"
    )(root, p_differentia_collision=0.5, mutate=True)
    assert leaf.blueberry == 0
    assert inner.blueberry == 0
    assert root.blueberry == 0


def test_assign_trie_origin_times_node_rank_mutate():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    leaf = impl.TrieLeafNode(parent=inner, taxon_label="A")
    root_ = hstrat.AssignOriginTimeNodeRankTriePostprocessor(
        assigned_property="blueberry"
    )(root, p_differentia_collision=0.5, mutate=False)
    assert not hasattr(leaf, "blueberry")
    assert not hasattr(inner, "blueberry")
    assert not hasattr(root, "blueberry")

    assert hasattr(root_, "blueberry")

    root = hstrat.AssignOriginTimeNodeRankTriePostprocessor(
        assigned_property="blueberry",
    )(root, p_differentia_collision=0.5, mutate=True)
    assert str(anytree.RenderTree(root).by_attr("blueberry")) == str(
        anytree.RenderTree(root_).by_attr("blueberry")
    )



import itertools as it
import os

import numpy as np
import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_find_mrca_id_asexual,
    alifestd_is_chronologically_ordered,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_fuzz(phylogeny_df: pd.DataFrame):
    assert alifestd_is_chronologically_ordered(phylogeny_df)

    for ids in it.islice(
        it.product(phylogeny_df["id"], phylogeny_df["id"]), 2
    ):
        res = alifestd_find_mrca_id_asexual(phylogeny_df, ids)
        assert res >= phylogeny_df["id"].min()
        assert all(res <= np.array(ids))

    for id_ in it.islice(reversed(phylogeny_df["id"]), 2):
        res = alifestd_find_mrca_id_asexual(phylogeny_df, [id_] * 3)
        assert res == id_


@pytest.mark.parametrize("mutate", [True, False])
def test_simple1(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 4],
            "ancestor_list": ["[None]", "[0]", "[1]", "[0]"],
        }
    )
    original_df = phylogeny_df.copy()

    res = alifestd_find_mrca_id_asexual(phylogeny_df, (0, 1, 2), mutate=mutate)
    assert res == 0

    res = alifestd_find_mrca_id_asexual(phylogeny_df, (0, 2), mutate=mutate)
    assert res == 0

    res = alifestd_find_mrca_id_asexual(phylogeny_df, (2,), mutate=mutate)
    assert res == 2

    res = alifestd_find_mrca_id_asexual(phylogeny_df, (2, 1), mutate=mutate)
    assert res == 1

    res = alifestd_find_mrca_id_asexual(phylogeny_df, (2, 4), mutate=mutate)
    assert res == 0

    if not mutate:
        assert original_df.equals(phylogeny_df)



from descend_template_phylogeny_ import (
    descend_template_phylogeny,
    descend_template_phylogeny_alifestd,
    descend_template_phylogeny_biopython,
    descend_template_phylogeny_dendropy,
    descend_template_phylogeny_naive,
    descend_template_phylogeny_networkx,
    descend_template_phylogeny_posthoc,
)
from generate_template_phylogeny import evolve_fitness_trait_population
from perfect_tracking import (
    DecantingPhyloTracker,
    GarbageCollectingPhyloTracker,
    PerfectBacktrackHandle,
    compile_perfect_backtrack_phylogeny,
    compile_phylogeny_from_lineage_iters,
    iter_perfect_backtrack_lineage,
)

from . import (
    descend_template_phylogeny_,
    generate_template_phylogeny,
    perfect_tracking,
)

__all__ = [
    "descend_template_phylogeny_",
    "generate_template_phylogeny",
    "perfect_tracking",
    "descend_template_phylogeny",
    "descend_template_phylogeny_alifestd",
    "descend_template_phylogeny_biopython",
    "descend_template_phylogeny_dendropy",
    "descend_template_phylogeny_naive",
    "descend_template_phylogeny_networkx",
    "descend_template_phylogeny_posthoc",
    "evolve_fitness_trait_population",
    "compile_perfect_backtrack_phylogeny",
    "compile_phylogeny_from_lineage_iters",
    "DecantingPhyloTracker",
    "GarbageCollectingPhyloTracker",
    "iter_perfect_backtrack_lineage",
    "PerfectBacktrackHandle",
]



import random

import pytest

from hstrat.hstrat import stochastic_algo


@pytest.mark.parametrize(
    "replicate",
    range(5),
)
def test_policy_consistency(replicate):
    random.seed(replicate)
    assert stochastic_algo.CalcRankAtColumnIndex is None



import typing

import opytional as opyt

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ._calc_rank_of_earliest_detectable_mrca_between import (
    calc_rank_of_earliest_detectable_mrca_between,
)


def calc_ranks_since_earliest_detectable_mrca_with(
    focal: HereditaryStratigraphicArtifact,
    other: HereditaryStratigraphicArtifact,
    confidence_level: float = 0.95,
) -> typing.Optional[int]:
    """How many generations have elapsed since the first where common ancestry
    with other could be detected?

    How many depositions have elapsed along focal's lineage since the earliest
    possible rank a MRCA between focal and other could be reliably detected at?

    Even if a true MRCA of focal and other exists, if it occured earlier
    than the rank calculated here it could not be reliably detected with
    sufficient confidence after accounting for the possibility of spurious
    differentia collisions. (Although subsequent spurious differentia
    collisions after the true MRCA of focal and other could lead to MRCA
    detection at such a rank.)

    Returns None if insufficient common ranks exist between focal and other
    to ever conclude at the given confidence level the existance of any
    common ancestry between focal and other (even if all strata at common
    ranks had equivalent differentiae).
    """
    return opyt.apply_if(
        calc_rank_of_earliest_detectable_mrca_between(
            focal,
            other,
            confidence_level=confidence_level,
        ),
        lambda rank: focal.GetNumStrataDeposited() - 1 - rank,
    )



import typing

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ._estimate_patristic_distance_between import (
    estimate_patristic_distance_between,
)


def ballpark_patristic_distance_between(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
) -> typing.Optional[float]:
    """Calculate a fast, rough estimate of the patristic distance between first
    and second.

    See `estimate_patristic_distance_between` for details.
    """
    return estimate_patristic_distance_between(
        first,
        second,
        estimator="maximum_likelihood",
        prior="arbitrary",
    )



import numpy as np
import pytest

from hstrat.hstrat import recency_proportional_resolution_curbed_algo


@pytest.mark.parametrize(
    "size_curb",
    [
        2,
        3,
        7,
        8,
        9,
        10,
        11,
        12,
        13,
        14,
        15,
        16,
        42,
        97,
        100,
        254,
        255,
        256,
        257,
    ],
)
@pytest.mark.parametrize(
    "time_sequence",
    [
        range(10**3),
        (i for i in range(10**2) for __ in range(2)),
        np.random.default_rng(1).integers(
            low=0,
            high=2**32,
            size=10,
        ),
        (2**32,),
    ],
)
def test_policy_consistency(size_curb, time_sequence):
    policy = recency_proportional_resolution_curbed_algo.Policy(size_curb)
    spec = policy.GetSpec()
    instance = recency_proportional_resolution_curbed_algo.GenDropRanks(spec)
    for num_strata_deposited in time_sequence:
        policy_requirement = {
            *policy.IterRetainedRanks(
                num_strata_deposited,
            )
        } - {
            *policy.IterRetainedRanks(
                num_strata_deposited + 1,
            )
        }
        for which in (
            instance,
            recency_proportional_resolution_curbed_algo.GenDropRanks(spec),
        ):
            assert sorted(
                which(
                    policy,
                    num_strata_deposited,
                    policy.IterRetainedRanks(num_strata_deposited),
                )
            ) == sorted(policy_requirement), (
                num_strata_deposited,
                size_curb,
                which,
            )



from ._Policy import Policy
from ._PolicySpec import PolicySpec
from ._enact._GenDropRanks import GenDropRanks
from ._enact._GenDropRanks_ import _GenDropRanks_impls
from ._invar._CalcMrcaUncertaintyAbsUpperBound import (
    CalcMrcaUncertaintyAbsUpperBound,
)
from ._invar._CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank import (
    CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank,
)
from ._invar._CalcMrcaUncertaintyAbsUpperBoundPessimalRank import (
    CalcMrcaUncertaintyAbsUpperBoundPessimalRank,
)
from ._invar._CalcMrcaUncertaintyRelUpperBound import (
    CalcMrcaUncertaintyRelUpperBound,
)
from ._invar._CalcMrcaUncertaintyRelUpperBoundAtPessimalRank import (
    CalcMrcaUncertaintyRelUpperBoundAtPessimalRank,
)
from ._invar._CalcMrcaUncertaintyRelUpperBoundPessimalRank import (
    CalcMrcaUncertaintyRelUpperBoundPessimalRank,
)
from ._invar._CalcNumStrataRetainedUpperBound import (
    CalcNumStrataRetainedUpperBound,
)
from ._scry._CalcMrcaUncertaintyAbsExact import CalcMrcaUncertaintyAbsExact
from ._scry._CalcMrcaUncertaintyRelExact import CalcMrcaUncertaintyRelExact
from ._scry._CalcNumStrataRetainedExact import CalcNumStrataRetainedExact
from ._scry._CalcRankAtColumnIndex import CalcRankAtColumnIndex
from ._scry._IterRetainedRanks import IterRetainedRanks

__all__ = [
    "CalcMrcaUncertaintyAbsExact",
    "CalcMrcaUncertaintyAbsUpperBound",
    "CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank",
    "CalcMrcaUncertaintyAbsUpperBoundPessimalRank",
    "CalcMrcaUncertaintyRelExact",
    "CalcMrcaUncertaintyRelUpperBound",
    "CalcMrcaUncertaintyRelUpperBoundAtPessimalRank",
    "CalcMrcaUncertaintyRelUpperBoundPessimalRank",
    "CalcNumStrataRetainedExact",
    "CalcNumStrataRetainedUpperBound",
    "CalcRankAtColumnIndex",
    "GenDropRanks",
    "_GenDropRanks_impls",
    "IterRetainedRanks",
    "Policy",
    "PolicySpec",
]



import typing

import anytree

from ._AnyTreeFastPreOrderIter import AnyTreeFastPreOrderIter


def AnyTreeFastLeafIter(
    node: anytree.AnyNode,
) -> typing.Iterator[anytree.AnyNode]:
    """Efficiently iterate over the leaf nodes of an AnyTree structure

    Parameters
    ----------
    node : anytree.AnyNode
        The root node of the tree (or subtree) for which the leaf nodes are to
        be iterated.

    Returns
    -------
    typing.Iterator[anytree.AnyNode]
        An iterator over leaf nodes of the tree (or subtree) rooted at `node`.
    """
    yield from filter(lambda x: x.is_leaf, AnyTreeFastPreOrderIter(node))



import types
import typing


def launder_impl_modules(
    all_: typing.List[types.ModuleType], name: str
) -> None:
    """Stamp name into __module__ of items to hide implementation detail.

    Important for making the documentation work nice.
    """
    for item in all_:
        try:
            item.__module__ = name
        except (AttributeError, TypeError):
            pass



import typing

import pandas as pd

from ._PerfectBacktrackHandle import PerfectBacktrackHandle
from ._compile_phylogeny_from_lineage_iters import (
    compile_phylogeny_from_lineage_iters,
)
from ._iter_perfect_backtrack_lineage import iter_perfect_backtrack_lineage


def compile_perfect_backtrack_phylogeny(
    population: typing.Iterable[PerfectBacktrackHandle],
) -> pd.DataFrame:
    """Compile phylogenetic history tracked using `PerfectBacktrackHandle`
    breadcrumbs.

    Parameters
    ----------
    population : iterable of `PerfectBactrackHandle`
        PerfectBacktrackHandle objects associated with extant population
        members.

    Returns
    -------
    pd.DataFrame
        Phylogenetic record in alife data standards format.
    """

    return compile_phylogeny_from_lineage_iters(
        iter_perfect_backtrack_lineage(extant_handle)
        for extant_handle in population
    )



from functools import lru_cache
import logging


# Keep track of n different messages and then log again
# adapted from https://stackoverflow.com/a/66062313
@lru_cache(1)
def log_once_in_a_row(msg: str):
    logger = logging.getLogger("hstrat")
    logger.info(msg)



import typing

import opytional as opyt

from ._detail import policy_param_focalizer_t


class MrcaUncertaintyRelExactEvaluator:
    """Enacts exact relative MRCA uncertainty parameterization requirement."""

    _at_num_strata_deposited: int
    _at_rank: typing.Optional[int]
    _policy_param_focalizer: policy_param_focalizer_t

    def __init__(
        self: "MrcaUncertaintyRelExactEvaluator",
        at_num_strata_deposited: int,
        at_rank: typing.Optional[int] = None,
        policy_param_focalizer: policy_param_focalizer_t = (
            lambda policy_t: lambda i: policy_t(i)
        ),
    ) -> None:
        """Initialize functor to evaluate exact relative MRCA uncertainty.

        Parameters
        ----------
        at_num_strata_deposited : int
            At what generation should we evaluate policy?
        at_rank : int, optional
            At what column position should we evaluate policy? If None, use
            pessimal position.
        policy_param_focalizer : callable
            Callable to create shim that constructs policy instance from
            parameter value. Default passes parameter value as sole argument to
            policy constructor.
        """
        self._at_num_strata_deposited = at_num_strata_deposited
        self._at_rank = at_rank
        self._policy_param_focalizer = policy_param_focalizer

    def __call__(
        self: "MrcaUncertaintyRelExactEvaluator",
        policy_t: typing.Type,
        parameter_value: int,
    ) -> float:
        """Get exact relative MRCA uncertainty under a specific parameter."""
        policy_factory = self._policy_param_focalizer(policy_t)
        policy = policy_factory(parameter_value)
        if self._at_rank is None:
            raise NotImplementedError(
                "CalcMrcaUncertaintyRelExactPessimalRank not yet implemented"
            )
        at_rank = opyt.or_value(
            self._at_rank,
            None,
            # policy.CalcMrcaUncertaintyRelExactPessimalRank(
            #     self._at_num_strata_deposited,
            #     self._at_num_strata_deposited,
            # ),
        )
        return policy.CalcMrcaUncertaintyRelExact(
            self._at_num_strata_deposited,
            self._at_num_strata_deposited,
            at_rank,
        )

    def __repr__(self: "MrcaUncertaintyRelExactEvaluator") -> str:
        return f"""{
            MrcaUncertaintyRelExactEvaluator.__qualname__
        }(at_num_strata_deposited={
            self._at_num_strata_deposited
        !r}, at_rank={
            self._at_rank
        !r}, policy_param_focalizer={
            self._policy_param_focalizer
        !r})"""

    def __str__(self: "MrcaUncertaintyRelExactEvaluator") -> str:
        title = "Exact Relative MRCA Uncertainty Evaluator"
        return f"""{
            title
        } (at num strata deposited: {
            self._at_num_strata_deposited
        }, at rank: {
            opyt.or_value(self._at_rank, 'pessimal')
        }, focalizer: {
            self._policy_param_focalizer
        })"""



import typing

from ..genome_instrumentation import HereditaryStratigraphicColumn
from ._col_from_records import col_from_records
from ._impl import col_records_from_pop_records


def pop_from_records(
    records: typing.Dict,
    progress_wrap: typing.Callable = lambda x: x,
    mutate: bool = False,
) -> typing.List[HereditaryStratigraphicColumn]:
    """Deserialize a sequence of `HereditaryStratigraphicColumn`s from a dict
    composed of builtin types.

    Parameters
    ----------
    records : dict
        Data to deserialize.
    progress_wrap : Callable, default identity function
        Wrapper applied around generation iterator and row generator for final
        phylogeny compilation process.

        Pass tqdm or equivalent to display progress bars.
    mutate : bool, default False
        Are side effects on the input argument `records` allowed?

    Returns
    -------
    population : List[HereditaryStratigraphicColumn]
        Deserialized population of HereditaryStratigraphicColumns.
    """

    col_records = col_records_from_pop_records(records, mutate=mutate)

    return [
        col_from_records(col_record)
        for col_record in progress_wrap(col_records)
    ]



"""Maintain constant size complexity and evenly-spaced strata.

The depth-proportional resolution policy ensures estimates of MRCA rank will
have uncertainty bounds less than or equal to a user-specified proportion of
the largest number of strata deposited on either column. Thus, MRCA rank
estimate uncertainty scales as O(n) with respect to the greater number of
strata deposited on either column.

Under the depth-proportional resolution policy, the number of strata retained
(i.e., space complexity) scales as O(1) with respect to the number of strata
deposited.

This retention policy guarantees that columns will retain appropriate strata so
that for any two columns with m and n strata deposited, the rank of the most
recent common ancestor can be determined with uncertainty of at most

    bound = floor(max(m, n) / depth_proportional_resolution)

ranks. Achieving this limit on uncertainty requires retaining sufficient strata
so that no more than bound ranks elapsed between any two strata. This policy
accumulates retained strata at a fixed interval until twice as many as
depth_proportional_resolution are at hand. Then, every other
retained stratum is purged and the cycle repeats with a new twice-as-wide
interval between retained strata.

Suppose depth_proportional_resolution is 3.

       guaranteed   actual
time   resolution   uncertainty   column
--------------------------------------------------------
1      0            0             |
2      0            0             ||
3      1            0             |||
4      1            0             ||||
5      1            0             |||||
6      2            2             | | ||
7      2            2             | | | |
8      2            2             | | | ||
9      3            2             | | | | |
10     3            2             | | | | ||
11     3            2             | | | | | |
12     4            4             |   |   |  |
13     4            4             |   |   |   |
14     4            4             |   |   |   ||
15     5            4             |   |   |   | |
16     5            4             |   |   |   |  |
17     5            4             |   |   |   |   |
18     6            4             |   |   |   |   ||
19     6            4             |   |   |   |   | |
20     6            4             |   |   |   |   |  |
21     7            4             |   |   |   |   |   |
22     7            4             |   |   |   |   |   ||
23     7            4             |   |   |   |   |   | |
24     8            8             |       |       |      |
25     8            8             |       |       |       |
26     8            8             |       |       |       ||
27     9            8             |       |       |       | |
28     9            8             |       |       |       |  |
29     9            8             |       |       |       |   |

See Also
--------
depth_proportional_resolution_tapered_algo:
    For a retention policy that achieves the same guarantees for depth-
    proportional resolution but purges unnecessary strata more graudally.
"""

from ...._auxiliary_lib import lazy_attach_stub

__getattr__, __dir__, __all__ = lazy_attach_stub(__name__, __file__)
del lazy_attach_stub



import numpy as np

from hstrat import hstrat


def test_estimate_rank_of_mrca_naive_no_common_ancestor():
    coincident_ranks = iter([0])
    p_differentia_collision = 0.5
    prior = object()

    result = hstrat.estimate_rank_of_mrca_naive(
        coincident_ranks, p_differentia_collision, prior
    )

    assert result is None


def test_estimate_rank_of_mrca_naive_with_common_ancestor1b():
    coincident_ranks = iter([1, 0])
    p_differentia_collision = 0.5

    result = hstrat.estimate_rank_of_mrca_naive(
        coincident_ranks, p_differentia_collision, hstrat.ArbitraryPrior()
    )

    assert result == 0


def test_estimate_rank_of_mrca_naive_with_common_ancestor1():
    coincident_ranks = iter([7, 3, 0])
    p_differentia_collision = 0.5
    prior = object()

    result = hstrat.estimate_rank_of_mrca_naive(
        coincident_ranks, p_differentia_collision, prior
    )

    assert isinstance(result, float)
    assert np.isclose(result, 4.5)


def test_estimate_rank_of_mrca_naive_with_common_ancestor2():
    coincident_ranks = [8, 7, 2, 0]
    p_differentia_collision = 0.5
    prior = object()

    result = hstrat.estimate_rank_of_mrca_naive(
        coincident_ranks, p_differentia_collision, prior
    )

    assert isinstance(result, float)
    assert np.isclose(result, 7)



import typing

import pandas as pd
import polars as pl

from ._coerce_to_pandas import _supported_iterables, _supported_mappings


def coerce_to_polars(obj: typing.Any, *, recurse: bool = False) -> typing.Any:
    """
    If a Pandas type is detected, coerce it to corresponding Polars type.
    """
    if isinstance(obj, (pd.Series, pd.DataFrame)):
        return pl.from_pandas(obj)
    elif recurse and isinstance(obj, _supported_iterables):
        return type(obj)(
            map(lambda x: coerce_to_polars(x, recurse=recurse), obj)
        )
    elif recurse and isinstance(obj, _supported_mappings):
        return {
            k: coerce_to_polars(v, recurse=recurse) for k, v in obj.items()
        }
    else:
        return obj



import typing


class PerfectBacktrackHandle:
    """Breadcrumb for perfectly backtracking a phylogenetic chain of descent.

    Because only backwards references are held, Python garbage collection takes
    care of pruning away extinct lineages.
    """

    # Achieves 60% the performance of C++ shared_ptr-based implementation in
    # tracking a unbranching linked list and outperforms the C++ implementation
    # by 30% in tracking a phylogenetic tree with random selection
    #
    #   p = [impl() for __ in range(100)]
    #   for __ in tqdm(range(10000000)):
    #       p[random.randrange(len(p))] = random.choice(p).CreateDescendant()
    #
    # see https://gist.github.com/mmore500/d6ee9011e38355f1be6c02a70db5b785
    # and https://gist.github.com/mmore500/71eb2f3f14afb02fba476dd2cea45210
    # (didn't complete debugging, but is representative of speed)

    # tried a dict-based database approach, but even with native implementaiton
    # was aboutt 3x slower than this approach
    # see https://gist.github.com/mmore500/7e9b85f005882d1e4b0206a9ef6d1de2

    parent: "PerfectBacktrackHandle"
    data: typing.Any

    __slots__ = ("parent", "data")

    def __init__(
        self: "PerfectBacktrackHandle",
        parent: "PerfectBacktrackHandle" = None,
        data: typing.Any = None,
    ) -> None:
        """Create backtracking breadcrumb, with `parent` as preceding
        breadcrumb in line of descent or `None` if associated with seed
        organism."""
        self.parent = parent
        self.data = data

    def CreateDescendant(
        self: "PerfectBacktrackHandle", data: typing.Any = None
    ) -> "PerfectBacktrackHandle":
        """Convenience factory method to create backtracking breadcrumbs
        succeeding `self` in line of descent."""
        cls = self.__class__
        return cls(self, data)



"""Tools to create phylogenies and corresponding extant populations of
hereditary stratigraph columns."""

from . import (
    descend_template_phylogeny_,
    generate_template_phylogeny,
    perfect_tracking,
)
from .._auxiliary_lib import lazy_attach

__getattr__, __dir__, __all__ = lazy_attach(
    __name__,
    submodules=[
        "descend_template_phylogeny_",
        "generate_template_phylogeny",
        "perfect_tracking",
    ],
    submod_attrs={
        "descend_template_phylogeny_": descend_template_phylogeny_.__all__,
        "generate_template_phylogeny": generate_template_phylogeny.__all__,
        "perfect_tracking": perfect_tracking.__all__,
    },
    should_launder=[].__contains__,
)
del lazy_attach



import typing

from ....._auxiliary_lib import bit_floor
from ._iter_target_recencies import iter_target_recencies


def iter_rank_seps(
    degree: int,
    interspersal: int,
    num_strata_deposited: int,
) -> typing.Iterator[int]:
    """Yield spacing between retained strata for each exponentially-spaced
    coverage target `pow` in ascending order.

    Yielded values will be powers of 2.
    """
    for target_recency in iter_target_recencies(degree, num_strata_deposited):
        # spacing between retained ranks
        target_retained_ranks_sep = max(
            target_recency / interspersal,
            1.0,
        )
        # round down to power of 2
        retained_ranks_sep = bit_floor(int(target_retained_ranks_sep))
        yield retained_ranks_sep



from hstrat._auxiliary_lib import count_trailing_ones


def test_no_trailing_ones():
    assert count_trailing_ones(0b0) == 0
    assert count_trailing_ones(-0b0) == 0
    assert count_trailing_ones(0b100) == 0
    assert count_trailing_ones(-0b100) == 0


def test_single_trailing_one():
    assert count_trailing_ones(0b10111101) == 1
    assert count_trailing_ones(-0b10111101) == 1
    assert count_trailing_ones(0b1) == 1
    assert count_trailing_ones(-0b1) == 1
    assert count_trailing_ones(0b101) == 1
    assert count_trailing_ones(-0b101) == 1


def test_multiple_trailing_ones():
    assert count_trailing_ones(0b11) == 2
    assert count_trailing_ones(-0b11) == 2
    assert count_trailing_ones(0b100011) == 2
    assert count_trailing_ones(-0b100011) == 2
    assert count_trailing_ones(0b1011) == 2
    assert count_trailing_ones(-0b1011) == 2
    assert count_trailing_ones(0b111) == 3
    assert count_trailing_ones(-0b111) == 3
    assert count_trailing_ones(0b100111) == 3
    assert count_trailing_ones(-0b100111) == 3
    assert count_trailing_ones(0b11111) == 5
    assert count_trailing_ones(-0b11111) == 5



from collections import Counter
import typing
import warnings

import numpy as np
import pandas as pd

from ._alifestd_assign_contiguous_ids import alifestd_assign_contiguous_ids
from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_is_asexual import alifestd_is_asexual
from ._alifestd_is_topologically_sorted import alifestd_is_topologically_sorted
from ._alifestd_make_ancestor_list_col import alifestd_make_ancestor_list_col
from ._alifestd_parse_ancestor_ids import alifestd_parse_ancestor_ids
from ._alifestd_topological_sort import alifestd_topological_sort
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col
from ._jit import jit
from ._jit_numpy_int64_t import jit_numpy_int64_t


@jit(nopython=True)
def _collapse_unifurcations(
    ancestor_ids: np.ndarray,
) -> typing.Tuple[np.ndarray, np.ndarray]:
    # assumes contiguous ids

    ref_counts = np.zeros(len(ancestor_ids), dtype=jit_numpy_int64_t)
    for ancestor_id in ancestor_ids:
        ref_counts[ancestor_id] += 1

    ids = np.arange(len(ancestor_ids))

    for pos, _ in enumerate(ancestor_ids):

        ancestor_id = ancestor_ids[pos]
        id_ = ids[pos]
        assert id_ == pos
        assert ancestor_id <= id_
        ref_count = ref_counts[pos]

        if ref_count == 1 and id_ != ancestor_id:
            # percolate ancestor over self
            ancestor_ids[pos] = ancestor_ids[ancestor_id]
            ids[pos] = ids[ancestor_id]
        else:
            # update referenced ancestor
            ancestor_ids[pos] = ids[ancestor_id]

    keep_filter = ref_counts != 1 | (ancestor_ids == ids)

    return keep_filter, ancestor_ids


def _alifestd_collapse_unifurcations_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool,
    root_ancestor_token: str,
) -> pd.DataFrame:
    """Optimized implementation for asexual phylogenies."""

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if not alifestd_is_topologically_sorted(phylogeny_df):
        phylogeny_df = alifestd_topological_sort(phylogeny_df, mutate=True)

    original_ids = phylogeny_df["id"].to_numpy().copy()
    if not alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df = alifestd_assign_contiguous_ids(
            phylogeny_df, mutate=True
        )

    keep_filter, ancestor_ids = _collapse_unifurcations(
        phylogeny_df["ancestor_id"].to_numpy()
    )
    drop_filter = ~keep_filter
    phylogeny_df.drop(
        phylogeny_df.index[drop_filter], inplace=True, axis="rows"
    )
    phylogeny_df.loc[:, "id"] = original_ids[keep_filter]
    phylogeny_df.loc[:, "ancestor_id"] = original_ids[
        ancestor_ids[keep_filter]
    ]
    phylogeny_df.loc[:, "ancestor_list"] = alifestd_make_ancestor_list_col(
        phylogeny_df["id"],
        phylogeny_df["ancestor_id"],
        root_ancestor_token=root_ancestor_token,
    )

    return phylogeny_df


def alifestd_collapse_unifurcations(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
    root_ancestor_token: str = "none",
) -> pd.DataFrame:
    """Pare record to bypass organisms with one ancestor and one descendant.

    The option `root_ancestor_token` will be sandwiched in brackets to create
    the ancestor list entry for genesis organisms. For example, the token
    "None" will yield the entry "[None]" and the token "" will yield the entry
    "[]". Default "none".

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if "branch_length" in phylogeny_df or "edge_length" in phylogeny_df:
        warnings.warn(
            "alifestd_collapse_unifurcations does not update branch length "
            "columns. Use `origin_time` to recalculate branch lengths for "
            "collapsed phylogeny."
        )

    # special optimized handling for asexual phylogenies
    if alifestd_is_asexual(phylogeny_df):
        return _alifestd_collapse_unifurcations_asexual(
            phylogeny_df,
            mutate=mutate,
            root_ancestor_token=root_ancestor_token,
        )

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    if not alifestd_is_topologically_sorted(phylogeny_df):
        phylogeny_df = alifestd_topological_sort(phylogeny_df, mutate=True)

    phylogeny_df.set_index("id", drop=False, inplace=True)

    ref_counts = Counter(
        id_
        for ancestor_list_str in phylogeny_df["ancestor_list"]
        for id_ in alifestd_parse_ancestor_ids(ancestor_list_str)
    )

    for id_ in phylogeny_df["id"]:
        ancestor_ids = alifestd_parse_ancestor_ids(
            phylogeny_df.loc[id_, "ancestor_list"]
        )
        if len(ancestor_ids) == 1 and ref_counts[id_] == 1:
            # percolate ancestor over self
            (ancestor_id,) = ancestor_ids
            phylogeny_df.loc[id_] = phylogeny_df.loc[ancestor_id]
        elif len(ancestor_ids):
            # update referenced ancestor
            phylogeny_df.loc[id_, "ancestor_list"] = str(
                [
                    int(phylogeny_df.loc[ancestor_id, "id"])
                    for ancestor_id in ancestor_ids
                ]
            )
        else:
            assert not ancestor_ids
            phylogeny_df.loc[id_, "ancestor_list"] = f"[{root_ancestor_token}]"

    assert "ancestor_id" not in phylogeny_df

    return phylogeny_df.drop_duplicates().reset_index(drop=True)



"""Tools to infer phylogenetic history from extant genome annotations."""

from . import estimators, pairwise, population, priors, tree
from .._auxiliary_lib import lazy_attach

__getattr__, __dir__, __all__ = lazy_attach(
    __name__,
    submodules=["estimators", "pairwise", "population", "priors", "tree"],
    submod_attrs={
        "estimators": estimators.__all__,
        "pairwise": pairwise.__all__,
        "population": population.__all__,
        "priors": priors.__all__,
        "tree": tree.__all__,
    },
    should_launder=[].__contains__,
)
